import logging

from PyQt5.QtCore import Qt, QUrl
from PyQt5.QtGui import QPixmap, QFont, QDesktopServices
from PyQt5.QtWidgets import (
    QWidget,
    QLabel,
    QPushButton,
    QHBoxLayout,
    QSpacerItem,
    QGroupBox,
    QTabWidget,
    QGridLayout,
)

from rare.components.tabs.store.shop_models import ShopGame
from rare.shared import LegendaryCoreSingleton
from rare.ui.components.tabs.store.shop_game_info import Ui_shop_info
from rare.utils.extra_widgets import ImageLabel
from rare.utils.misc import icon
from rare.widgets.loading_widget import LoadingWidget

logger = logging.getLogger("ShopInfo")


class ShopGameInfo(QWidget, Ui_shop_info):
    game: ShopGame
    data: dict

    # TODO Design
    def __init__(self, installed_titles: list, api_core):
        super(ShopGameInfo, self).__init__()
        self.setupUi(self)
        self.core = LegendaryCoreSingleton()
        self.api_core = api_core
        self.installed = installed_titles
        self.open_store_button.clicked.connect(self.button_clicked)
        self.image = ImageLabel()
        self.image_stack.addWidget(self.image)
        self.image_stack.addWidget(LoadingWidget())
        warn_label = QLabel()
        warn_label.setPixmap(
            icon("fa.warning").pixmap(160, 160).scaled(240, 320, Qt.IgnoreAspectRatio)
        )
        self.image_stack.addWidget(warn_label)

        self.wishlist_button.clicked.connect(self.add_to_wishlist)
        self.in_wishlist = False
        self.wishlist = []

    def handle_wishlist_update(self, data):
        if data and data[0] == "error":
            return
        self.wishlist = [i["offer"]["title"] for i in data]
        if self.title_str in self.wishlist:
            self.in_wishlist = True
            self.wishlist_button.setVisible(True)
            self.wishlist_button.setText(self.tr("Remove from Wishlist"))
        else:
            self.in_wishlist = False
            self.wishlist_button.setVisible(False)

    def update_game(self, data: dict):
        self.image_stack.setCurrentIndex(1)
        self.title.setText(data["title"])
        self.title_str = data["title"]
        self.api_core.get_wishlist(self.handle_wishlist_update)
        for i in reversed(range(self.req_group_box.layout().count())):
            self.req_group_box.layout().itemAt(i).widget().deleteLater()
        slug = data["productSlug"]
        if not slug:
            for mapping in data["offerMappings"]:
                if mapping["pageType"] == "productHome":
                    slug = mapping["pageSlug"]
                    break
            else:
                logger.error("Could not get page information")
                slug = ""
        if "/home" in slug:
            slug = slug.replace("/home", "")
        self.slug = slug

        if data["namespace"] in self.installed:
            self.open_store_button.setText(self.tr("Show Game on Epic Page"))
            self.owned_label.setVisible(True)
        else:
            self.open_store_button.setText(self.tr("Buy Game in Epic Games Store"))
            self.owned_label.setVisible(False)

        for i in range(self.req_group_box.layout().count()):
            self.req_group_box.layout().itemAt(i).widget().deleteLater()

        self.price.setText(self.tr("Loading"))
        self.wishlist_button.setVisible(False)
        # self.title.setText(self.tr("Loading"))
        self.image.setPixmap(QPixmap())
        self.data = data
        is_bundle = False
        for i in data["categories"]:
            if "bundles" in i.get("path", ""):
                is_bundle = True

        # init API request
        if slug:
            self.api_core.get_game(slug, is_bundle, self.data_received)
        else:
            self.data_received({})

    def add_to_wishlist(self):
        if not self.in_wishlist:
            return
            # self.api_core.add_to_wishlist(self.game.namespace, self.game.offer_id,
            #                             lambda success: self.wishlist_button.setText(self.tr("Remove from wishlist"))
            #                              if success else self.wishlist_button.setText("Something goes wrong"))
        else:
            self.api_core.remove_from_wishlist(
                self.game.namespace,
                self.game.offer_id,
                lambda success: self.wishlist_button.setVisible(False)
                if success
                else self.wishlist_button.setText("Something goes wrong"),
            )

    def data_received(self, game):
        try:
            self.game = ShopGame.from_json(game, self.data)
        except Exception as e:
            logger.error(str(e))
            self.price.setText("Error")
            self.req_group_box.setVisible(False)
            for img in self.data.get("keyImages"):
                if img["type"] in [
                    "DieselStoreFrontWide",
                    "OfferImageTall",
                    "VaultClosed",
                    "ProductLogo",
                ]:
                    self.image.update_image(img["url"], self.title_str, size=(240, 320))
                    self.image_stack.setCurrentIndex(0)
                    break
            else:
                self.image_stack.setCurrentIndex(2)
            self.price.setText("")
            self.discount_price.setText("")
            self.social_link_gb.setVisible(False)
            self.tags.setText("")
            self.dev.setText(self.data.get("seller", {}).get("name", ""))
            return
        self.title.setText(self.game.title)

        self.price.setFont(QFont())
        if self.game.price == "0" or self.game.price == 0:
            self.price.setText(self.tr("Free"))
        else:
            self.price.setText(self.game.price)
        if self.game.price != self.game.discount_price:
            font = QFont()
            font.setStrikeOut(True)
            self.price.setFont(font)
            self.discount_price.setText(
                self.game.discount_price
                if self.game.discount_price != "0"
                else self.tr("Free")
            )
            self.discount_price.setVisible(True)
        else:
            self.discount_price.setVisible(False)

        bold_font = QFont()
        bold_font.setBold(True)

        if self.game.reqs:
            req_tabs = QTabWidget()
            for system in self.game.reqs:
                min_label = QLabel(self.tr("Minimum"))
                min_label.setFont(bold_font)
                rec_label = QLabel(self.tr("Recommend"))
                rec_label.setFont(bold_font)
                req_widget = QWidget()
                req_widget.setLayout(QGridLayout())
                req_widget.layout().addWidget(min_label, 0, 1)
                req_widget.layout().addWidget(rec_label, 0, 2)
                for i, (key, value) in enumerate(
                        self.game.reqs.get(system, {}).items()
                ):
                    req_widget.layout().addWidget(QLabel(key), i + 1, 0)
                    min_label = QLabel(value[0])
                    min_label.setWordWrap(True)
                    req_widget.layout().addWidget(min_label, i + 1, 1)
                    rec_label = QLabel(value[1])
                    rec_label.setWordWrap(True)
                    req_widget.layout().addWidget(rec_label, i + 1, 2)
                req_tabs.addTab(req_widget, system)
            self.req_group_box.layout().addWidget(req_tabs)
        else:
            self.req_group_box.layout().addWidget(
                QLabel(self.tr("Could not get requirements"))
            )
        self.req_group_box.setVisible(True)
        if self.game.image_urls.front_tall:
            img_url = self.game.image_urls.front_tall
        elif self.game.image_urls.offer_image_tall:
            img_url = self.game.image_urls.offer_image_tall
        elif self.game.image_urls.product_logo:
            img_url = self.game.image_urls.product_logo
        else:
            img_url = ""
        self.image.update_image(img_url, self.game.title, (240, 320))

        self.image_stack.setCurrentIndex(0)
        try:
            if isinstance(self.game.developer, list):
                self.dev.setText(", ".join(self.game.developer))
            else:
                self.dev.setText(self.game.developer)
        except KeyError:
            pass
        self.tags.setText(", ".join(self.game.tags))

        # clear Layout
        for widget in (
                self.social_link_gb.layout().itemAt(i)
                for i in range(self.social_link_gb.layout().count())
        ):
            if not isinstance(widget, QSpacerItem):
                widget.widget().deleteLater()
        self.social_link_gb.deleteLater()
        self.social_link_gb = QGroupBox(self.tr("Social Links"))
        self.social_link_gb.setLayout(QHBoxLayout())

        self.layout().insertWidget(3, self.social_link_gb)

        self.social_link_gb.layout().addStretch(1)
        link_count = 0
        for name, url in self.game.links:

            if name.lower() == "homepage":
                icn = icon("mdi.web", "fa.search", scale_factor=1.5)
            else:
                try:
                    icn = icon(f"mdi.{name.lower()}", f"fa.{name.lower()}", scale_factor=1.5)
                except Exception as e:
                    logger.error(str(e))
                    continue

            button = SocialButton(icn, url)
            self.social_link_gb.layout().addWidget(button)
            link_count += 1
            self.social_link_gb.layout().addStretch(1)

        if link_count == 0:
            self.social_link_gb.setVisible(False)
        else:
            self.social_link_gb.setVisible(True)
        self.social_link_gb.layout().addStretch(1)

    def add_wishlist_items(self, wishlist):
        wishlist = wishlist["data"]["Wishlist"]["wishlistItems"]["elements"]
        for game in wishlist:
            self.wishlist.append(game["offer"]["title"])

    def button_clicked(self):
        QDesktopServices.openUrl(QUrl(f"https://www.epicgames.com/store/{self.core.language_code}/p/{self.slug}"))


class SocialButton(QPushButton):
    def __init__(self, icn, url):
        super(SocialButton, self).__init__(icn, "")
        self.url = url
        self.clicked.connect(lambda: QDesktopServices.openUrl(QUrl(url)))
        self.setToolTip(url)



from logging import getLogger

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QPixmap, QImage, QMovie
from PyQt5.QtWidgets import (
    QStyle,
    QLabel,
    QHBoxLayout,
    QWidget,
    QPushButton,
    QLineEdit,
    QToolButton,
)

from rare.utils.misc import icon as qta_icon
from rare.utils.paths import cache_dir
from rare.utils.qt_requests import QtRequests

logger = getLogger("ExtraWidgets")


class WaitingSpinner(QLabel):
    def __init__(self, parent=None):
        super(WaitingSpinner, self).__init__(parent=parent)
        self.setObjectName(type(self).__name__)
        self.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        self.movie = QMovie(":/images/loader.gif")
        self.setMovie(self.movie)
        self.movie.start()


class SelectViewWidget(QWidget):
    toggled = pyqtSignal()

    def __init__(self, icon_view: bool, parent=None):
        super(SelectViewWidget, self).__init__(parent=parent)
        self.icon_view = icon_view
        self.icon_button = QPushButton(self)
        self.icon_button.setObjectName(f"{type(self).__name__}Button")
        self.list_button = QPushButton(self)
        self.list_button.setObjectName(f"{type(self).__name__}Button")

        if icon_view:
            self.icon_button.setIcon(qta_icon("mdi.view-grid-outline", "ei.th-large", color="orange"))
            self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list"))
        else:
            self.icon_button.setIcon(qta_icon("mdi.view-grid-outline", "ei.th-large"))
            self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list", color="orange"))

        self.icon_button.clicked.connect(self.icon)
        self.list_button.clicked.connect(self.list)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.icon_button)
        layout.addWidget(self.list_button)

        self.setLayout(layout)

    def isChecked(self):
        return self.icon_view

    def icon(self):
        self.icon_button.setIcon(qta_icon("mdi.view-grid-outline", "ei.th-large", color="orange"))
        self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list"))
        self.icon_view = False
        self.toggled.emit()

    def list(self):
        self.icon_button.setIcon(qta_icon("mdi.view-grid-outline", "ei.th-large"))
        self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list", color="orange"))
        self.icon_view = True
        self.toggled.emit()


class ImageLabel(QLabel):
    image = None
    img_size = None
    name = ""

    def __init__(self, parent=None):
        super(ImageLabel, self).__init__(parent=parent)
        self.manager = QtRequests(
            cache=str(cache_dir().joinpath("store")),
            parent=self
        )

    def update_image(self, url, name="", size: tuple = (240, 320)):
        self.setFixedSize(*size)
        self.img_size = size
        self.name = name
        for c in r'<>?":|\/* ':
            self.name = self.name.replace(c, "")
        if self.img_size[0] > self.img_size[1]:
            name_extension = "wide"
        else:
            name_extension = "tall"
        self.name = f"{self.name}_{name_extension}.png"
        self.manager.get(url, self.image_ready)

    def image_ready(self, data):
        try:
            self.setPixmap(QPixmap())
        except Exception:
            logger.warning("C++ object already removed, when image ready")
            return
        image = QImage()
        image.loadFromData(data)
        image = image.scaled(
            *self.img_size[:2],
            Qt.KeepAspectRatio,
            transformMode=Qt.SmoothTransformation,
        )

        pixmap = QPixmap().fromImage(image)
        self.setPixmap(pixmap)


class ButtonLineEdit(QLineEdit):
    buttonClicked = pyqtSignal()

    def __init__(self, icon_name, placeholder_text: str, parent=None):
        super(ButtonLineEdit, self).__init__(parent=parent)
        self.setObjectName(type(self).__name__)

        self.button = QToolButton(self)
        self.button.setObjectName(f"{type(self).__name__}Button")
        self.button.setIcon(qta_icon(icon_name, color="white"))
        self.button.setStyleSheet(
            f"QToolButton#{self.button.objectName()} {{border: 0px; padding: 0px;}}"
        )
        self.button.setCursor(Qt.ArrowCursor)
        self.button.clicked.connect(self.buttonClicked.emit)
        self.setPlaceholderText(placeholder_text)
        frame_width = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        button_size = self.button.sizeHint()

        self.setStyleSheet(
            f"QLineEdit#{self.objectName()} {{padding-right: {(button_size.width() + frame_width + 1)}px; }}"
        )
        self.setMinimumSize(
            max(self.minimumSizeHint().width(), button_size.width() + frame_width * 2 + 2),
            max(
                self.minimumSizeHint().height(),
                button_size.height() + frame_width * 2 + 2,
            ),
        )

    def resizeEvent(self, event):
        frame_width = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        button_size = self.button.sizeHint()
        self.button.move(
            self.rect().right() - frame_width - button_size.width(),
            (self.rect().bottom() - button_size.height() + 1) // 2,
        )
        super(ButtonLineEdit, self).resizeEvent(event)



import re
import sys
from collections import ChainMap
from typing import Any, Union

from PyQt5.QtCore import Qt, QModelIndex, QAbstractTableModel, pyqtSlot
from PyQt5.QtGui import QFont

from rare.lgndr.core import LegendaryCore
from rare.utils.misc import icon


class EnvVarsTableModel(QAbstractTableModel):
    def __init__(self, core: LegendaryCore, parent = None):
        super(EnvVarsTableModel, self).__init__(parent=parent)
        self.core = core

        # lk: validator matches anything starting with a letter or underscore
        # lk: and containing letters, numbers or underscores.
        # lk: Empty strings are considered invalid.
        self.__validator = re.compile(r"(^[A-Za-z_][A-Za-z0-9_]*)")
        self.__data_map: ChainMap = ChainMap()

        self.__readonly = [
            "STEAM_COMPAT_DATA_PATH",
            "STEAM_COMPAT_CLIENT_INSTALL_PATH",
            "WINEPREFIX",
            "DXVK_HUD",
            "MANGOHUD_CONFIG",
        ]

        self.__default: str = "default"
        self.__appname: str = None

    @pyqtSlot(str)
    def reset(self):
        if not self.__appname:
            return
        self.load(self.__appname)

    def load(self, app_name: str):
        self.__appname = app_name
        self.beginResetModel()
        if not self.core.lgd.config.has_section(f"{self.__appname}.env"):
            self.core.lgd.config[f"{self.__appname}.env"] = {}
        self.__data_map = ChainMap(
            self.core.lgd.config[f"{self.__appname}.env"],
            self.core.lgd.config[f"{self.__default}.env"] if self.__appname != self.__default else {}
        )
        self.endResetModel()

    def __key(self, index: Union[QModelIndex, int]):
        if isinstance(index, QModelIndex):
            index = index.row()
        try:
            return list(self.__data_map)[index]
        except Exception as e:
            return ""

    def __is_local(self, index: Union[QModelIndex, int]):
        key = self.__key(index)
        return key in self.__data_map.maps[0].keys()

    def __is_global(self, index: Union[QModelIndex, int]):
        key = self.__key(index)
        return key in self.__data_map.maps[1].keys()

    def __is_readonly(self, index: Union[QModelIndex, int]):
        key = self.__key(index)
        return key in self.__readonly

    def __value(self, index: Union[QModelIndex, int]):
        if isinstance(index, QModelIndex):
            index = index.row()
        return self.__data_map[self.__key(index)]

    def __title(self, section: int):
        if section == 0:
            return self.tr("Key")
        elif section == 1:
            return self.tr("Value")

    def __data_length(self):
        return len(self.__data_map)

    def __is_key_valid(self, value: str):
        match = self.__validator.match(value)
        if not match:
            return False
        return value == match.group(0)

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Any:
        if role == Qt.DisplayRole or role == Qt.EditRole:
            if index.row() == self.__data_length():
                return ""
            if index.column() == 0:
                return self.__key(index)
            else:
                return self.__value(index)

        if role == Qt.TextAlignmentRole:
            if index.column() == 0:
                return Qt.AlignVCenter + Qt.AlignRight
            else:
                return Qt.AlignVCenter + Qt.AlignLeft

        if role == Qt.FontRole:
            font = QFont("Monospace")
            font.setStyleHint(QFont.Monospace)
            if index.row() < self.__data_length() and not self.__is_local(index):
                font.setWeight(QFont.Bold)
            else:
                font.setWeight(QFont.Normal)
            return font

        if role == Qt.ToolTipRole:
            if index.row() == self.__data_length():
                if index.column() == 1:
                    return self.tr("Disabled, please set the variable name first.")
                else:
                    return None
            if self.__key(index) in self.__readonly:
                return self.tr("Readonly, please edit this via setting the appropriate setting.")

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.DisplayRole) -> Any:
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return self.__title(section)
        if role == Qt.DecorationRole:
            if orientation == Qt.Vertical:
                if section < self.__data_length():
                    if self.__is_readonly(section) or not self.__is_local(section):
                        return icon("mdi.lock", "ei.lock")
                    if self.__is_global(section) and self.__is_local(section):
                        return icon("mdi.refresh", "ei.refresh")
                    if self.__is_local(section):
                        return icon("mdi.delete", "ei.remove-sign")
                return icon("mdi.plus", "ei.plus-sign")
        if role == Qt.TextAlignmentRole:
            return Qt.AlignVCenter + Qt.AlignHCenter
        return None

    def flags(self, index: QModelIndex):
        # Disable the value cell on rows without a name
        if index.row() == self.__data_length():
            if index.column() == 1:
                return super().flags(index) ^ Qt.ItemIsEnabled
            else:
                return Qt.ItemIsEditable | super().flags(index)

        # Disable readonly variables
        if self.__is_readonly(index):
            return super().flags(index) ^ Qt.ItemIsEnabled

        return Qt.ItemIsEditable | super().flags(index)

    def rowCount(self, parent: QModelIndex = None) -> int:
        parent = parent if parent else QModelIndex()
        # The length of the outer list.
        return self.__data_length() + 1

    def columnCount(self, parent: QModelIndex = None) -> int:
        parent = parent if parent else QModelIndex()
        return 2

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.DisplayRole) -> bool:
        if role != Qt.EditRole:
            return False

        if index.column() == 0:
            # lk: for what is considered valid input, look at `__validator`
            if (not self.__is_key_valid(value)) or value in self.__readonly:
                return False
            # Do not accept existing variable names (this also protects against unchanged contents)
            if value in self.__data_map.keys():
                return False

            if index.row() == self.__data_length():
                self.beginInsertRows(QModelIndex(), self.rowCount(index), self.rowCount(index))
                self.endInsertRows()
                self.__data_map[value] = ""
                self.core.lgd.save_config()
                self.dataChanged.emit(index, index, [])
                self.headerDataChanged.emit(Qt.Vertical, index.row(), index.row())
                # if we are on the last row, add a new last row to the table when setting the variable name
            else:
                # if we are not in the last row, we have to update an existing variable name
                old_key = self.__key(index)
                self.__data_map[value] = self.__data_map[old_key]
                self.core.lgd.save_config()
                # since we delete and add a new key, the new key will be moved at the end
                # deleting a local key can have the following effects
                # unique local key:
                #   old key deleted, new key added -> update range from index to end
                # old local key masking global key:
                #   old key remains, new key added -> insert row for new key, update from index to end
                # new key masking global key:
                #   can't happen because we do not accept existing keys
                if old_key in self.__data_map.maps[0].keys():
                    # delete the old key if it is a local one, replacing a local key
                    del self.__data_map[old_key]
                    self.core.lgd.save_config()
                if old_key in self.__data_map.maps[1].keys():
                    self.beginInsertRows(QModelIndex(), self.__data_length(), self.__data_length())
                    self.endInsertRows()
                self.dataChanged.emit(index, self.index(index.row(), 1), [])
                self.dataChanged.emit(self.index(self.__data_length() - 1, 0), self.index(self.__data_length() - 1, 1), [])
                self.headerDataChanged.emit(Qt.Vertical, index.row(), index.row())
                self.headerDataChanged.emit(Qt.Vertical, self.__data_length() - 1, self.__data_length() - 1)

        else:
            # lk: the check for key existance before assigning a value is ommitted
            # lk: The `Value` field of the table is disabled if the `Name`/key field is empty
            # lk: making is "impossible" to assign a value to an empty key
            # Do not accept unchanged contents
            if value == self.__value(index):
                return False

            self.__data_map[self.__key(index)] = value
            self.core.lgd.save_config()
            self.dataChanged.emit(self.index(index.row(), 0), index, [])
            self.headerDataChanged.emit(Qt.Vertical, index.row(), index.row())
        return True

    def removeRow(self, row: int, parent: QModelIndex = None) -> bool:
        parent = parent if parent else QModelIndex()
        # Refuse to remove the last row
        if row == self.__data_length():
            return False
        # Refuse to remove readonly keys
        if self.__is_readonly(row):
            return False
        # Refuse to remove keys that don't have our app_name (in this case the default)
        if not self.__is_local(row):
            return False
        if self.__is_global(row):
            del self.__data_map[self.__key(row)]
            self.core.lgd.save_config()
            self.dataChanged.emit(self.index(row, 0), self.index(row, 1), [])
            self.headerDataChanged.emit(Qt.Vertical, row, row)
        else:
            self.beginRemoveRows(QModelIndex(), row, row)
            del self.__data_map[self.__key(row)]
            self.core.lgd.save_config()
            self.endRemoveRows()
        return True


if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication, QDialog, QVBoxLayout, QTableView, QHeaderView

    from rare.resources import static_css
    from rare.resources.stylesheets import RareStyle
    from rare.utils.misc import set_style_sheet
    from legendary.core import LegendaryCore

    class MainDialog(QDialog):
        def __init__(self):
            super().__init__()

            self.table = QTableView()
            self.model = EnvVarsTableModel(LegendaryCore())
            self.model.load("Tamarind")

            self.table.setModel(self.model)
            self.table.verticalHeader().sectionPressed.disconnect()
            self.table.horizontalHeader().sectionPressed.disconnect()
            self.table.verticalHeader().sectionClicked.connect(self.model.removeRow)
            self.table.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
            self.table.horizontalHeader().setStretchLastSection(True)
            self.table.setCornerButtonEnabled(False)

            self.setLayout(QVBoxLayout(self))
            self.layout().addWidget(self.table)

    app = QApplication(sys.argv)

    set_style_sheet("RareStyle")

    window = MainDialog()
    window.setFixedSize(800, 600)
    window.show()
    app.exec()



import logging
import os
import platform
import sys
import time
import traceback
from argparse import Namespace

import legendary
from PyQt5.QtCore import QSettings, QTranslator, QT_VERSION_STR, PYQT_VERSION_STR, QObject, pyqtSignal, pyqtSlot, Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication, QMessageBox

import rare.resources.resources
from rare.utils import paths
from rare.utils.misc import set_color_pallete, set_style_sheet, get_static_style


class RareAppException(QObject):
    exception = pyqtSignal(object, object, object)

    def __init__(self, parent=None):
        super(RareAppException, self).__init__(parent=parent)
        self.logger = logging.getLogger(type(self).__name__)
        sys.excepthook = self._excepthook
        self.exception.connect(self._on_exception)

    def _excepthook(self, exc_type: object, exc_value: object, exc_tb: object):
        self.exception.emit(exc_type, exc_value, exc_tb)

    def _handler(self, exc_type, exc_value, exc_tb) -> bool:
        return False

    @pyqtSlot(object, object, object)
    def _on_exception(self, exc_type, exc_value, exc_tb):
        message = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
        if self._handler(exc_type, exc_value, exc_tb):
            return
        self.logger.fatal(message)
        action = QMessageBox.warning(
            None, exc_type.__name__, message,
            buttons=QMessageBox.Ignore | QMessageBox.Close,
            defaultButton=QMessageBox.Ignore
        )
        if action == QMessageBox.RejectRole:
            QApplication.exit(1)


class RareApp(QApplication):
    def __init__(self, args: Namespace, log_file: str):
        super(RareApp, self).__init__(sys.argv)
        self.logger = logging.getLogger(type(self).__name__)
        self._hook = RareAppException(self)
        self.setQuitOnLastWindowClosed(False)
        self.setAttribute(Qt.AA_DontUseNativeDialogs, True)

        self.setApplicationName("Rare")
        self.setOrganizationName("Rare")

        # Create directories after QStandardPaths has been initialized
        paths.create_dirs()

        # Clean any existing logging handlers from library imports
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        start_time = time.strftime("%y-%m-%d--%H-%M")  # year-month-day-hour-minute
        file_handler = logging.FileHandler(
            filename=os.path.join(paths.log_dir(), log_file.format(start_time)),
            encoding="utf-8",
        )
        file_handler.setFormatter(fmt=logging.Formatter("[%(name)s] %(levelname)s: %(message)s"))

        # Set up common logging channel to stderr
        if args.debug:
            logging.basicConfig(
                format="[%(name)s] %(levelname)s: %(message)s",
                level=logging.DEBUG,
                stream=sys.stderr,
            )
            file_handler.setLevel(logging.DEBUG)
            logging.root.addHandler(file_handler)
            logging.getLogger().setLevel(level=logging.DEBUG)
            # keep requests, asyncio and pillow quiet
            logging.getLogger("requests").setLevel(logging.WARNING)
            logging.getLogger("urllib3").setLevel(logging.WARNING)
            logging.getLogger("asyncio").setLevel(logging.WARNING)
            self.logger.info(
                f"Launching Rare version {rare.__version__} Codename: {rare.__codename__}\n"
                f" - Using Legendary {legendary.__version__} Codename: {legendary.__codename__} as backend\n"
                f" - Operating System: {platform.system()}, Python version: {platform.python_version()}\n"
                f" - Running {sys.executable} {' '.join(sys.argv)}\n"
                f" - Qt version: {QT_VERSION_STR}, PyQt version: {PYQT_VERSION_STR}"
            )
        else:
            logging.basicConfig(
                format="[%(name)s] %(levelname)s: %(message)s",
                level=logging.INFO,
                stream=sys.stderr,
            )
            file_handler.setLevel(logging.DEBUG)
            logging.root.addHandler(file_handler)
            self.logger.info(f"Launching Rare version {rare.__version__}")
            self.logger.info(f"Operating System: {platform.system()}")

        self.settings = QSettings()

        # Translator
        self.translator = QTranslator()
        self.qt_translator = QTranslator()

        # Style
        # lk: this is a bit silly but serves well until we have a class
        # lk: store the default qt style name from the system on startup as a property for later reference
        self.setProperty("rareDefaultQtStyle", self.style().objectName())
        if (
                self.settings.value("color_scheme", None) is None
                and self.settings.value("style_sheet", None) is None
        ):
            self.settings.setValue("color_scheme", "")
            self.settings.setValue("style_sheet", "RareStyle")

        if color_scheme := self.settings.value("color_scheme", False):
            self.settings.setValue("style_sheet", "")
            set_color_pallete(color_scheme)
        elif style_sheet := self.settings.value("style_sheet", False):
            self.settings.setValue("color_scheme", "")
            set_style_sheet(style_sheet)
        else:
            self.setStyleSheet(get_static_style())
        self.setWindowIcon(QIcon(":/images/Rare.png"))

    def load_translator(self, lang: str):
        if os.path.isfile(f := os.path.join(paths.resources_path, "languages", f"{lang}.qm")):
            self.translator.load(f)
            self.logger.info(f"Your language is supported: {lang}")
        elif not lang == "en":
            self.logger.info("Your language is not supported")
        self.installTranslator(self.translator)

        # translator for qt stuff
        if os.path.isfile(f := os.path.join(paths.resources_path, f"qt_{lang}.qm")):
            self.qt_translator = QTranslator()
            self.qt_translator.load(f)
            self.installTranslator(self.qt_translator)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/settings/game_settings.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_GameSettings(object):
    def setupUi(self, GameSettings):
        GameSettings.setObjectName("GameSettings")
        GameSettings.resize(505, 261)
        self.game_settings_layout = QtWidgets.QVBoxLayout(GameSettings)
        self.game_settings_layout.setObjectName("game_settings_layout")
        self.launch_settings_group = QtWidgets.QGroupBox(GameSettings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.launch_settings_group.sizePolicy().hasHeightForWidth())
        self.launch_settings_group.setSizePolicy(sizePolicy)
        self.launch_settings_group.setObjectName("launch_settings_group")
        self.launch_settings_layout = QtWidgets.QFormLayout(self.launch_settings_group)
        self.launch_settings_layout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.launch_settings_layout.setObjectName("launch_settings_layout")
        self.skip_update_label = QtWidgets.QLabel(self.launch_settings_group)
        self.skip_update_label.setObjectName("skip_update_label")
        self.launch_settings_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.skip_update_label)
        self.skip_update = QtWidgets.QComboBox(self.launch_settings_group)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.skip_update.sizePolicy().hasHeightForWidth())
        self.skip_update.setSizePolicy(sizePolicy)
        self.skip_update.setObjectName("skip_update")
        self.skip_update.addItem("")
        self.skip_update.addItem("")
        self.skip_update.addItem("")
        self.launch_settings_layout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.skip_update)
        self.offline_label = QtWidgets.QLabel(self.launch_settings_group)
        self.offline_label.setObjectName("offline_label")
        self.launch_settings_layout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.offline_label)
        self.offline = QtWidgets.QComboBox(self.launch_settings_group)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.offline.sizePolicy().hasHeightForWidth())
        self.offline.setSizePolicy(sizePolicy)
        self.offline.setObjectName("offline")
        self.offline.addItem("")
        self.offline.addItem("")
        self.offline.addItem("")
        self.launch_settings_layout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.offline)
        self.launch_params_label = QtWidgets.QLabel(self.launch_settings_group)
        self.launch_params_label.setObjectName("launch_params_label")
        self.launch_settings_layout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.launch_params_label)
        self.launch_params = QtWidgets.QLineEdit(self.launch_settings_group)
        self.launch_params.setObjectName("launch_params")
        self.launch_settings_layout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.launch_params)
        self.game_settings_layout.addWidget(self.launch_settings_group)
        self.proton_layout = QtWidgets.QVBoxLayout()
        self.proton_layout.setObjectName("proton_layout")
        self.game_settings_layout.addLayout(self.proton_layout)
        self.linux_settings_widget = QtWidgets.QWidget(GameSettings)
        self.linux_settings_widget.setObjectName("linux_settings_widget")
        self.linux_settings_layout = QtWidgets.QVBoxLayout(self.linux_settings_widget)
        self.linux_settings_layout.setContentsMargins(0, 0, 0, 0)
        self.linux_settings_layout.setObjectName("linux_settings_layout")
        self.game_settings_layout.addWidget(self.linux_settings_widget)

        self.retranslateUi(GameSettings)

    def retranslateUi(self, GameSettings):
        _translate = QtCore.QCoreApplication.translate
        GameSettings.setWindowTitle(_translate("GameSettings", "GameSettings"))
        self.launch_settings_group.setTitle(_translate("GameSettings", "Launch Settings"))
        self.skip_update_label.setText(_translate("GameSettings", "Skip update check"))
        self.skip_update.setItemText(0, _translate("GameSettings", "Default"))
        self.skip_update.setItemText(1, _translate("GameSettings", "Yes"))
        self.skip_update.setItemText(2, _translate("GameSettings", "No"))
        self.offline_label.setText(_translate("GameSettings", "Offline mode"))
        self.offline.setItemText(0, _translate("GameSettings", "Default"))
        self.offline.setItemText(1, _translate("GameSettings", "Yes"))
        self.offline.setItemText(2, _translate("GameSettings", "No"))
        self.launch_params_label.setText(_translate("GameSettings", "Launch parameters"))
        self.launch_params.setPlaceholderText(_translate("GameSettings", "parameters"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    GameSettings = QtWidgets.QWidget()
    ui = Ui_GameSettings()
    ui.setupUi(GameSettings)
    GameSettings.show()
    sys.exit(app.exec_())



import os
import shutil
from argparse import Namespace
from datetime import datetime, timezone
from typing import Optional

import requests.exceptions
from PyQt5.QtCore import QThreadPool, QTimer, pyqtSlot, Qt
from PyQt5.QtWidgets import QApplication, QMessageBox
from requests import HTTPError

from rare.components.dialogs.launch_dialog import LaunchDialog
from rare.components.main_window import MainWindow
from rare.shared import RareCore
from rare.utils import config_helper, paths
from rare.utils.misc import ExitCodes
from rare.widgets.rare_app import RareApp, RareAppException


class RareException(RareAppException):
    def __init__(self, parent=None):
        super(RareException, self).__init__(parent=parent)

    def _handler(self, exc_type, exc_value, exc_tb) -> bool:
        if exc_type == HTTPError:
            try:
                if RareCore.instance() is not None:
                    if RareCore.instance().core().login():
                        return True
                raise ValueError
            except Exception as e:
                self.logger.fatal(str(e))
                QMessageBox.warning(None, "Error", self.tr("Failed to login"))
                QApplication.exit(1)
        return False


class Rare(RareApp):
    def __init__(self, args: Namespace):
        super(Rare, self).__init__(args, f"{type(self).__name__}_{{0}}.log")
        self._hook.deleteLater()
        self._hook = RareException(self)
        self.rcore = RareCore(args=args)
        self.args = RareCore.instance().args()
        self.signals = RareCore.instance().signals()
        self.core = RareCore.instance().core()

        lang = self.settings.value("language", self.core.language_code, type=str)
        self.load_translator(lang)

        # set Application name for settings
        self.main_window: Optional[MainWindow] = None
        self.launch_dialog: Optional[LaunchDialog] = None
        self.timer: Optional[QTimer] = None

        # This launches the application after it has been instantiated.
        # The timer's signal will be serviced once we call `exec()` on the application
        QTimer.singleShot(0, self.launch_app)

    def poke_timer(self):
        dt_exp = datetime.fromisoformat(self.core.lgd.userdata['expires_at'][:-1]).replace(tzinfo=timezone.utc)
        dt_now = datetime.utcnow().replace(tzinfo=timezone.utc)
        td = abs(dt_exp - dt_now)
        self.timer.start(int(td.total_seconds() - 60) * 1000)
        self.logger.info(f"Renewed session expires at {self.core.lgd.userdata['expires_at']}")

    def re_login(self):
        self.logger.info("Session expires shortly. Renew session")
        try:
            self.core.login()
        except requests.exceptions.ConnectionError:
            self.timer.start(3000)  # try again if no connection
            return
        self.poke_timer()

    @pyqtSlot()
    def launch_app(self):
        self.launch_dialog = LaunchDialog(parent=None)
        self.launch_dialog.rejected.connect(self.__on_exit_app)
        # lk: the reason we use the `start_app` signal here instead of accepted, is to keep the dialog
        # until the main window has been created, then we call `accept()` to close the dialog
        self.launch_dialog.start_app.connect(self.__on_start_app)
        self.launch_dialog.start_app.connect(self.launch_dialog.accept)
        self.launch_dialog.login()

    @pyqtSlot()
    def __on_start_app(self):
        self.timer = QTimer()
        self.timer.timeout.connect(self.re_login)
        self.poke_timer()

        self.main_window = MainWindow()
        self.main_window.exit_app.connect(self.__on_exit_app)

        if not self.args.silent:
            self.main_window.show()

        if self.args.test_start:
            self.main_window.close()
            self.main_window = None
            self.__on_exit_app(0)

    @pyqtSlot()
    @pyqtSlot(int)
    def __on_exit_app(self, exit_code=0):
        threadpool = QThreadPool.globalInstance()
        threadpool.waitForDone()
        if self.timer is not None:
            self.timer.stop()
            self.timer.deleteLater()
            self.timer = None
        self.rcore.deleteLater()
        del self.rcore
        self.processEvents()
        shutil.rmtree(paths.tmp_dir())
        os.makedirs(paths.tmp_dir())

        self.exit(exit_code)


def start(args) -> int:
    while True:
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
        app = Rare(args)
        exit_code = app.exec()
        del app
        if exit_code != ExitCodes.LOGOUT:
            break
    return exit_code



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/extra/console_env.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ConsoleEnv(object):
    def setupUi(self, ConsoleEnv):
        ConsoleEnv.setObjectName("ConsoleEnv")
        ConsoleEnv.resize(600, 400)
        self.layout = QtWidgets.QVBoxLayout(ConsoleEnv)
        self.layout.setObjectName("layout")
        self.table = QtWidgets.QTableWidget(ConsoleEnv)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        self.table.setFont(font)
        self.table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.table.setAlternatingRowColors(True)
        self.table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.table.setCornerButtonEnabled(True)
        self.table.setColumnCount(2)
        self.table.setObjectName("table")
        self.table.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.table.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.table.setHorizontalHeaderItem(1, item)
        self.table.horizontalHeader().setVisible(True)
        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.verticalHeader().setVisible(False)
        self.layout.addWidget(self.table)
        self.buttons = QtWidgets.QDialogButtonBox(ConsoleEnv)
        self.buttons.setOrientation(QtCore.Qt.Horizontal)
        self.buttons.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttons.setObjectName("buttons")
        self.layout.addWidget(self.buttons)

        self.retranslateUi(ConsoleEnv)
        self.buttons.accepted.connect(ConsoleEnv.accept) # type: ignore
        self.buttons.rejected.connect(ConsoleEnv.reject) # type: ignore

    def retranslateUi(self, ConsoleEnv):
        _translate = QtCore.QCoreApplication.translate
        ConsoleEnv.setWindowTitle(_translate("ConsoleEnv", "Rare - Console Environment"))
        self.table.setSortingEnabled(True)
        item = self.table.horizontalHeaderItem(0)
        item.setText(_translate("ConsoleEnv", "Variable"))
        item = self.table.horizontalHeaderItem(1)
        item.setText(_translate("ConsoleEnv", "Value"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    ConsoleEnv = QtWidgets.QDialog()
    ui = Ui_ConsoleEnv()
    ui.setupUi(ConsoleEnv)
    ConsoleEnv.show()
    sys.exit(app.exec_())



import os
from typing import Union, List

from legendary.core import LegendaryCore


class PathSpec:
    __egl_path_vars = {
        "{appdata}": os.path.expandvars("%LOCALAPPDATA%"),
        "{userdir}": os.path.expandvars("%USERPROFILE%/Documents"),
        # '{userprofile}': os.path.expandvars('%userprofile%'),  # possibly wrong
        "{usersavedgames}": os.path.expandvars("%USERPROFILE%/Saved Games"),
    }
    egl_appdata: str = r"%LOCALAPPDATA%\EpicGamesLauncher\Saved\Config\Windows"
    egl_programdata: str = r"%PROGRAMDATA%\Epic\EpicGamesLauncher\Data\Manifests"
    wine_programdata: str = r"dosdevices/c:/ProgramData"

    def __init__(self, core: LegendaryCore = None, app_name: str = "default"):
        if core is not None:
            self.__egl_path_vars.update({"{epicid}": core.lgd.userdata["account_id"]})
        self.app_name = app_name

    def cook(self, path: str) -> str:
        cooked_path = [self.__egl_path_vars.get(p.lower(), p) for p in path.split("/")]
        return os.path.join(*cooked_path)

    @property
    def wine_egl_programdata(self):
        return self.egl_programdata.replace("\\", "/").replace("%PROGRAMDATA%", self.wine_programdata)

    def wine_egl_prefixes(self, results: int = 0) -> Union[List[str], str]:
        possible_prefixes = [
            os.path.expanduser("~/.wine"),
            os.path.expanduser("~/Games/epic-games-store"),
        ]
        prefixes = []
        for prefix in possible_prefixes:
            if os.path.exists(os.path.join(prefix, self.wine_egl_programdata)):
                prefixes.append(prefix)
        if not prefixes:
            return str()
        if not results:
            return prefixes
        elif results == 1:
            return prefixes[0]
        else:
            return prefixes[:results]



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/settings/widgets/wrapper.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_WrapperSettings(object):
    def setupUi(self, WrapperSettings):
        WrapperSettings.setObjectName("WrapperSettings")
        WrapperSettings.resize(199, 30)
        WrapperSettings.setWindowTitle("WrapperSettings")
        self.wrapper_settings_layout = QtWidgets.QHBoxLayout(WrapperSettings)
        self.wrapper_settings_layout.setContentsMargins(0, 0, 0, 0)
        self.wrapper_settings_layout.setObjectName("wrapper_settings_layout")
        self.widget_stack = QtWidgets.QStackedWidget(WrapperSettings)
        self.widget_stack.setObjectName("widget_stack")
        self.label_page = QtWidgets.QWidget()
        self.label_page.setObjectName("label_page")
        self.label_page_layout = QtWidgets.QVBoxLayout(self.label_page)
        self.label_page_layout.setContentsMargins(0, 0, 0, 0)
        self.label_page_layout.setObjectName("label_page_layout")
        self.no_wrapper_label = QtWidgets.QLabel(self.label_page)
        self.no_wrapper_label.setObjectName("no_wrapper_label")
        self.label_page_layout.addWidget(self.no_wrapper_label)
        self.widget_stack.addWidget(self.label_page)
        self.wrapper_settings_layout.addWidget(self.widget_stack)
        self.add_button = QtWidgets.QPushButton(WrapperSettings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.add_button.sizePolicy().hasHeightForWidth())
        self.add_button.setSizePolicy(sizePolicy)
        self.add_button.setObjectName("add_button")
        self.wrapper_settings_layout.addWidget(self.add_button, 0, QtCore.Qt.AlignTop)

        self.retranslateUi(WrapperSettings)
        self.widget_stack.setCurrentIndex(0)

    def retranslateUi(self, WrapperSettings):
        _translate = QtCore.QCoreApplication.translate
        self.no_wrapper_label.setText(_translate("WrapperSettings", "No wrapper added"))
        self.add_button.setText(_translate("WrapperSettings", "Add wrapper"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    WrapperSettings = QtWidgets.QWidget()
    ui = Ui_WrapperSettings()
    ui.setupUi(WrapperSettings)
    WrapperSettings.show()
    sys.exit(app.exec_())



from typing import Optional

from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QKeyEvent
from PyQt5.QtWidgets import QTreeView

from rare.models.game import RareGame
from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton, ArgumentsSingleton
from rare.utils.json_formatter import QJsonModel
from rare.widgets.side_tab import SideTabWidget, SideTabContents
from .game_dlc import GameDlc
from .game_info import GameInfo
from .game_settings import GameSettings
from .cloud_saves import CloudSaves


class GameInfoTabs(SideTabWidget):
    # str: app_name
    import_clicked = pyqtSignal(str)

    def __init__(self, parent=None):
        super(GameInfoTabs, self).__init__(show_back=True, parent=parent)
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()
        self.args = ArgumentsSingleton()

        self.info_tab = GameInfo(self)
        self.info_tab.import_clicked.connect(self.import_clicked)
        self.info_index = self.addTab(self.info_tab, self.tr("Information"))

        self.settings_tab = GameSettings(self)
        self.settings_index = self.addTab(self.settings_tab, self.tr("Settings"))

        self.cloud_saves_tab = CloudSaves(self)
        self.cloud_saves_index = self.addTab(self.cloud_saves_tab, self.tr("Cloud Saves"))

        self.dlc_tab = GameDlc(self)
        self.dlc_index = self.addTab(self.dlc_tab, self.tr("Downloadable Content"))

        # FIXME: Hiding didn't work, so don't add these tabs in normal mode. Fix this properly later
        if self.args.debug:
            self.game_meta_view = GameMetadataView(self)
            self.game_meta_index = self.addTab(self.game_meta_view, self.tr("Game Metadata"))
            self.igame_meta_view = GameMetadataView(self)
            self.igame_meta_index = self.addTab(self.igame_meta_view, self.tr("InstalledGame Metadata"))

        self.setCurrentIndex(self.info_index)

    def update_game(self, rgame: RareGame):
        self.info_tab.update_game(rgame)

        self.settings_tab.load_settings(rgame)
        self.settings_tab.setEnabled(rgame.is_installed or rgame.is_origin)

        self.dlc_tab.update_dlcs(rgame)
        self.dlc_tab.setEnabled(rgame.is_installed and bool(rgame.owned_dlcs))

        self.cloud_saves_tab.update_game(rgame)
        # self.cloud_saves_tab.setEnabled(rgame.game.supports_cloud_saves or rgame.game.supports_mac_cloud_saves)

        if self.args.debug:
            self.game_meta_view.update_game(rgame, rgame.game)
            self.igame_meta_view.update_game(rgame, rgame.igame)

        self.setCurrentIndex(self.info_index)

    def keyPressEvent(self, e: QKeyEvent):
        if e.key() == Qt.Key_Escape:
            self.back_clicked.emit()


class GameMetadataView(QTreeView, SideTabContents):
    def __init__(self, parent=None):
        super(GameMetadataView, self).__init__(parent=parent)
        self.implements_scrollarea = True
        self.setColumnWidth(0, 300)
        self.setWordWrap(True)
        self.model = QJsonModel()
        self.setModel(self.model)
        self.rgame: Optional[RareGame] = None

    def update_game(self, rgame: RareGame, view):
        self.rgame = rgame
        self.set_title.emit(self.rgame.app_title)
        self.model.clear()
        try:
            self.model.load(vars(view))
        except Exception as e:
            pass
        self.resizeColumnToContents(0)



import logging
# import warnings


class LgndrException(RuntimeError):
    def __init__(self, message="Error in Legendary"):
        self.message = message
        super(LgndrException, self).__init__(self.message)


class LgndrWarning(RuntimeWarning):
    def __init__(self, message="Warning in Legendary"):
        self.message = message
        super(LgndrWarning, self).__init__(self.message)


# Minimum exception levels per class to get back useful error strings
# CLI: ERROR
# Core: CRITICAL (FATAL)
class LgndrLogHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level=level)

    def emit(self, record: logging.LogRecord) -> None:
        if record.levelno >= self.level:
            raise LgndrException(record.getMessage())
        # if self.level > record.levelno >= logging.WARNING:
        #     warnings.warn(record.getMessage())



import os
import sys
from typing import Union, Type

import qstylizer.style
from PyQt5.QtCore import QDir, QObject
from PyQt5.QtGui import QColor
from PyQt5.pyrcc import RCCResourceLibrary, CONSTANT_COMPRESSLEVEL_DEFAULT, CONSTANT_COMPRESSTHRESHOLD_DEFAULT
from PyQt5.sip import wrappertype

from rare.utils.misc import widget_object_name

verbose = True
compressLevel = 6
compressThreshold = CONSTANT_COMPRESSTHRESHOLD_DEFAULT
resourceRoot = ''


def processResourceFile(filenamesIn, filenameOut, listFiles):
    if verbose:
        sys.stderr.write("PyQt5 resource compiler\n")

    # Setup.
    library = RCCResourceLibrary()
    library.setInputFiles(filenamesIn)
    library.setVerbose(verbose)
    library.setCompressLevel(compressLevel)
    library.setCompressThreshold(compressThreshold)
    library.setResourceRoot(resourceRoot)

    if not library.readFiles():
        return False

    if filenameOut == '-':
        filenameOut = ''

    if listFiles:
        # Open the output file or use stdout if not specified.
        if filenameOut:
            try:
                out_fd = open(filenameOut, 'w')
            except Exception:
                sys.stderr.write(
                        "Unable to open %s for writing\n" % filenameOut)
                return False
        else:
            out_fd = sys.stdout

        for df in library.dataFiles():
            out_fd.write("%s\n" % QDir.cleanPath(df))

        if out_fd is not sys.stdout:
            out_fd.close()

        return True

    return library.output(filenameOut)


def css_name(widget: Union[wrappertype, QObject, Type], subwidget: str = ""):
    return f"#{widget_object_name(widget, '')}{subwidget}"


css = qstylizer.style.StyleSheet()


# [Un]InstallButton
css.QPushButton["#InstallButton"].setValues(
    borderColor=QColor(0, 180, 0).name(),
    backgroundColor=QColor(0, 120, 0).name()
)
css.QPushButton["#InstallButton"].hover.setValues(
    borderColor=QColor(0, 135, 0).name(),
    backgroundColor=QColor(0, 90, 0).name()
)
css.QPushButton["#InstallButton"].disabled.setValues(
    borderColor=QColor(0, 60, 0).name(),
    backgroundColor=QColor(0, 40, 0).name()
)
css.QPushButton["#UninstallButton"].setValues(
    borderColor=QColor(180, 0, 0).name(),
    backgroundColor=QColor(120, 0, 0).name()
)
css.QPushButton["#UninstallButton"].hover.setValues(
    borderColor=QColor(135, 0, 0).name(),
    backgroundColor=QColor(90, 0, 0).name()
)
css.QPushButton["#UninstallButton"].disabled.setValues(
    borderColor=QColor(60, 0, 0).name(),
    backgroundColor=QColor(40, 0, 0).name()
)


# QueueWorkerLabel
css.QLabel["#QueueWorkerLabel"].setValues(
    borderWidth="1px",
    borderStyle="solid",
    borderRadius="3px",
)
verify_color = QColor("#d6af57")
css.QLabel["#QueueWorkerLabel"]['[workerType="Verify"]'].setValues(
    borderColor=verify_color.darker(200).name(),
    backgroundColor=verify_color.darker(400).name()
)
move_color = QColor("#41cad9")
css.QLabel["#QueueWorkerLabel"]['[workerType="Move"]'].setValues(
    borderColor=move_color.darker(200).name(),
    backgroundColor=move_color.darker(400).name()
)


# ProgressLabel
from rare.components.tabs.games.game_widgets.library_widget import ProgressLabel
css.QLabel[css_name(ProgressLabel)].setValues(
    borderWidth="1px",
    borderRadius="5%",
    fontWeight="bold",
    fontSize="16pt",
)


# IconGameWidget
from rare.components.tabs.games.game_widgets.icon_widget import IconWidget
icon_background_props = {
    "backgroundColor": "rgba(0, 0, 0, 65%)",
}
css.QLabel[css_name(IconWidget, "StatusLabel")].setValues(
    fontWeight="bold",
    color="white",
    **icon_background_props,
    borderRadius="5%",
    borderTopLeftRadius="11%",
    borderTopRightRadius="11%",
)
css.QWidget[css_name(IconWidget, "MiniWidget")].setValues(
    color="rgb(238, 238, 238)",
    **icon_background_props,
    borderRadius="5%",
    borderBottomLeftRadius="9%",
    borderBottomRightRadius="9%",
)
icon_bottom_label_props = {
    "color": "white",
    "backgroundColor": "rgba(0, 0, 0, 0%)",
}
css.QLabel[css_name(IconWidget, "TitleLabel")].setValues(
    fontWeight="bold",
    **icon_bottom_label_props,
)
css.QLabel[css_name(IconWidget, "TooltipLabel")].setValues(
    **icon_bottom_label_props,
)
icon_square_button_props = {
    "border": "1px solid black",
    "borderRadius": "10%",
}
icon_square_button_props.update(icon_background_props)
css.QPushButton[css_name(IconWidget, "Button")].setValues(
    **icon_square_button_props
)
css.QPushButton[css_name(IconWidget, "Button")].hover.borderColor.setValue("gray")


# ListGameWidget
from rare.components.tabs.games.game_widgets.list_widget import ListWidget
css.QLabel[css_name(ListWidget,"TitleLabel")].fontWeight.setValue("bold")
list_status_label_props = {
    "color": "white",
    "backgroundColor": "rgba(0, 0, 0, 75%)",
    "border": "1px solid black",
    "borderRadius": "5px",
}
css.QLabel[css_name(ListWidget, "StatusLabel")].setValues(
    fontWeight="bold",
    **list_status_label_props,
)
css.QLabel[css_name(ListWidget, "TooltipLabel")].setValues(
    **list_status_label_props,
)
css.QPushButton[css_name(ListWidget, "Button")].textAlign.setValue("left")
css.QLabel[css_name(ListWidget, "InfoLabel")].color.setValue("#999")


# WaitingSpinner
from rare.utils.extra_widgets import WaitingSpinner
css.QLabel[css_name(WaitingSpinner)].setValues(
    marginLeft="auto",
    marginRight="auto",
)


# SelectViewWidget
from rare.utils.extra_widgets import SelectViewWidget
css.QPushButton[css_name(SelectViewWidget, "Button")].setValues(
    border="none",
    backgroundColor="transparent",
)


if __name__ == "__main__":
    with open("stylesheet.qss", "w") as qss:
        qss.write(f'\n/* This file is auto-generated from "{os.path.basename(__file__)}". DO NOT EDIT!!! */\n\n')
        qss.write(css.toString())

    if not processResourceFile(["stylesheet.qrc"], "__init__.py", False):
        print("Error while creating compiled resources")
        sys.exit(1)

    sys.exit(0)



from PyQt5.QtCore import QSize
from PyQt5.QtWidgets import QTabBar, QToolButton

from rare.utils.misc import icon


class MainTabBar(QTabBar):
    def __init__(self, parent=None):
        super(MainTabBar, self).__init__(parent=parent)
        self.setObjectName("MainTabBar")
        font = self.font()
        font.setPointSize(font.pointSize() + 2)
        font.setBold(True)
        self.setFont(font)
        self.expanded = -1

    def tabSizeHint(self, index):
        size = super(MainTabBar, self).tabSizeHint(index)
        if index == self.expanded:
            offset = self.width()
            for index in range(self.count()):
                offset -= super(MainTabBar, self).tabSizeHint(index).width()
            size.setWidth(max(size.width(), size.width() + offset))
        return size


class TabButtonWidget(QToolButton):
    def __init__(self, button_icon: str, tool_tip: str, fallback_icon=None, parent=None):
        super(TabButtonWidget, self).__init__(parent=parent)
        self.setText("Icon")
        self.setPopupMode(QToolButton.InstantPopup)
        self.setIcon(icon(button_icon, fallback_icon, scale_factor=1.25))
        self.setToolTip(tool_tip)
        self.setIconSize(QSize(25, 25))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/dialogs/login/login_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_LoginDialog(object):
    def setupUi(self, LoginDialog):
        LoginDialog.setObjectName("LoginDialog")
        LoginDialog.resize(241, 128)
        self.main_layout = QtWidgets.QVBoxLayout(LoginDialog)
        self.main_layout.setObjectName("main_layout")
        spacerItem = QtWidgets.QSpacerItem(0, 17, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.main_layout.addItem(spacerItem)
        self.welcome_label = QtWidgets.QLabel(LoginDialog)
        self.welcome_label.setObjectName("welcome_label")
        self.main_layout.addWidget(self.welcome_label)
        spacerItem1 = QtWidgets.QSpacerItem(0, 17, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.main_layout.addItem(spacerItem1)
        self.login_stack_layout = QtWidgets.QVBoxLayout()
        self.login_stack_layout.setObjectName("login_stack_layout")
        self.main_layout.addLayout(self.login_stack_layout)
        self.button_layout = QtWidgets.QHBoxLayout()
        self.button_layout.setObjectName("button_layout")
        self.exit_button = QtWidgets.QPushButton(LoginDialog)
        self.exit_button.setObjectName("exit_button")
        self.button_layout.addWidget(self.exit_button)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.button_layout.addItem(spacerItem2)
        self.back_button = QtWidgets.QPushButton(LoginDialog)
        self.back_button.setObjectName("back_button")
        self.button_layout.addWidget(self.back_button)
        self.next_button = QtWidgets.QPushButton(LoginDialog)
        self.next_button.setObjectName("next_button")
        self.button_layout.addWidget(self.next_button)
        self.main_layout.addLayout(self.button_layout)

        self.retranslateUi(LoginDialog)

    def retranslateUi(self, LoginDialog):
        _translate = QtCore.QCoreApplication.translate
        LoginDialog.setWindowTitle(_translate("LoginDialog", "Login"))
        self.welcome_label.setText(_translate("LoginDialog", "<h1>Welcome to Rare</h1>"))
        self.exit_button.setText(_translate("LoginDialog", "Exit"))
        self.back_button.setText(_translate("LoginDialog", "Back"))
        self.next_button.setText(_translate("LoginDialog", "Next"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    LoginDialog = QtWidgets.QDialog()
    ui = Ui_LoginDialog()
    ui.setupUi(LoginDialog)
    LoginDialog.show()
    sys.exit(app.exec_())



import datetime
from dataclasses import dataclass


class ImageUrlModel:
    def __init__(
        self,
        front_tall: str = "",
        offer_image_tall: str = "",
        thumbnail: str = "",
        front_wide: str = "",
        offer_image_wide: str = "",
        product_logo: str = "",
    ):
        self.front_tall = front_tall
        self.offer_image_tall = offer_image_tall
        self.thumbnail = thumbnail
        self.front_wide = front_wide
        self.offer_image_wide = offer_image_wide
        self.product_logo = product_logo

    @classmethod
    def from_json(cls, json_data: list):
        tmp = cls()
        for item in json_data:
            if item["type"] == "Thumbnail":
                tmp.thumbnail = item["url"]
            elif item["type"] == "DieselStoreFrontTall":
                tmp.front_tall = item["url"]
            elif item["type"] == "DieselStoreFrontWide":
                tmp.front_wide = item["url"]
            elif item["type"] == "OfferImageTall":
                tmp.offer_image_tall = item["url"]
            elif item["type"] == "OfferImageWide":
                tmp.offer_image_wide = item["url"]
            elif item["type"] == "ProductLogo":
                tmp.product_logo = item["url"]
        return tmp


class ShopGame:
    # TODO: Copyrights etc
    def __init__(
        self,
        title: str = "",
        image_urls: ImageUrlModel = None,
        social_links: dict = None,
        langs: list = None,
        reqs: dict = None,
        publisher: str = "",
        developer: str = "",
        original_price: str = "",
        discount_price: str = "",
        tags: list = None,
        namespace: str = "",
        offer_id: str = "",
    ):
        self.title = title
        self.image_urls = image_urls
        self.links = []
        if social_links:
            for item in social_links:
                if item.startswith("link"):
                    self.links.append(
                        tuple((item.replace("link", ""), social_links[item]))
                    )
        else:
            self.links = []
        self.languages = langs
        self.reqs = reqs
        self.publisher = publisher
        self.developer = developer
        self.price = original_price
        self.discount_price = discount_price
        self.tags = tags
        self.namespace = namespace
        self.offer_id = offer_id

    @classmethod
    def from_json(cls, api_data: dict, search_data: dict):
        if isinstance(api_data, list):
            for product in api_data:
                if product["_title"] == "home":
                    api_data = product
                    break
        if "pages" in api_data.keys():
            for page in api_data["pages"]:
                if page["_slug"] == "home":
                    api_data = page
                    break
        tmp = cls()
        tmp.title = search_data.get("title", "Fail")
        tmp.image_urls = ImageUrlModel.from_json(search_data["keyImages"])
        links = api_data["data"]["socialLinks"]
        tmp.links = []
        for item in links:
            if item.startswith("link"):
                tmp.links.append(tuple((item.replace("link", ""), links[item])))
        tmp.available_voice_langs = api_data["data"]["requirements"].get(
            "languages", "Failed"
        )
        tmp.reqs = {}
        for i, system in enumerate(api_data["data"]["requirements"].get("systems", [])):
            try:
                tmp.reqs[system["systemType"]] = {}
            except KeyError:
                continue
            for req in system["details"]:
                try:
                    tmp.reqs[system["systemType"]][req["title"]] = (
                        req["minimum"],
                        req["recommended"],
                    )
                except KeyError:
                    pass
        tmp.publisher = api_data["data"]["meta"].get("publisher", "")
        tmp.developer = api_data["data"]["meta"].get("developer", "")
        if not tmp.developer:
            for i in search_data["customAttributes"]:
                if i["key"] == "developerName":
                    tmp.developer = i["value"]
        tmp.price = search_data["price"]["totalPrice"]["fmtPrice"]["originalPrice"]
        tmp.discount_price = search_data["price"]["totalPrice"]["fmtPrice"][
            "discountPrice"
        ]
        tmp.tags = [
            i.replace("_", " ").capitalize()
            for i in api_data["data"]["meta"].get("tags", [])
        ]
        tmp.namespace = search_data["namespace"]
        tmp.offer_id = search_data["id"]

        return tmp


@dataclass
class BrowseModel:
    category: str = "games/edition/base|bundles/games|editors|software/edition/base"
    count: int = 30
    language_code: str = "en"
    country_code: str = "US"
    keywords: str = ""
    sortDir: str = "DESC"
    start: int = 0
    tag: str = ""
    withMapping: bool = True
    withPrice: bool = True
    date: str = (
        f"[,{datetime.datetime.strftime(datetime.datetime.now(), '%Y-%m-%dT%H:%M:%S')}.420Z]"
    )
    price: str = ""
    onSale: bool = False

    @property
    def __dict__(self):
        payload = {
            "allowCountries": self.country_code,
            "category": self.category,
            "count": self.count,
            "country": self.country_code,
            "keywords": self.keywords,
            "locale": self.language_code,
            "priceRange": self.price,
            "releaseDate": self.date,
            "sortBy": "releaseDate",
            "sortDir": self.sortDir,
            "start": self.start,
            "tag": self.tag,
            "withPrice": self.withPrice,
        }
        if self.price == "free":
            payload["freeGame"] = True
            payload.pop("priceRange")
        elif self.price.startswith("<price>"):
            payload["priceRange"] = self.price.replace("<price>", "")
        if self.onSale:
            payload["onSale"] = True

        if self.price:
            payload["effectiveDate"] = self.date
        else:
            payload.pop("priceRange")

        return payload



from PyQt5.QtCore import (
    pyqtSlot,
    QEvent,
    QEasingCurve,
    QParallelAnimationGroup,
    QAbstractAnimation,
    QPropertyAnimation,
    Qt,
    QPoint,
)
from PyQt5.QtWidgets import QStackedWidget, QGestureEvent, QSwipeGesture


class SlidingStackedWidget(QStackedWidget):
    """
    Taken from: https://stackoverflow.com/a/52597972
    """

    def __init__(self, parent=None):
        super(SlidingStackedWidget, self).__init__(parent)

        self.m_direction = Qt.Horizontal
        self.m_speed = 500
        self.m_animationtype = QEasingCurve.OutBack
        self.m_now = 0
        self.m_next = 0
        self.m_wrap = False
        self.m_pnow = QPoint(0, 0)
        self.m_active = False

    def setDirection(self, direction: Qt.Orientation) -> None:
        self.m_direction = direction

    def setSpeed(self, speed: int) -> None:
        self.m_speed = speed

    def setAnimation(self, animationtype: QEasingCurve.Type) -> None:
        self.m_animationtype = animationtype

    def setWrap(self, wrap: bool) -> None:
        self.m_wrap = wrap

    @pyqtSlot()
    def slideInPrev(self):
        now = self.currentIndex()
        if self.m_wrap or now > 0:
            self.slideInIndex(now - 1)

    @pyqtSlot()
    def slideInNext(self):
        now = self.currentIndex()
        if self.m_wrap or now < (self.count() - 1):
            self.slideInIndex(now + 1)

    def slideInIndex(self, idx):
        if idx > (self.count() - 1):
            idx = idx % self.count()
        elif idx < 0:
            idx = (idx + self.count()) % self.count()
        self.slideInWidget(self.widget(idx))

    def slideInWidget(self, newwidget):
        if self.m_active:
            return

        self.m_active = True

        _now = self.currentIndex()
        _next = self.indexOf(newwidget)

        if _now == _next:
            self.m_active = False
            return

        offsetx, offsety = self.frameRect().width(), self.frameRect().height()
        self.widget(_next).setGeometry(self.frameRect())

        if not self.m_direction == Qt.Horizontal:
            if _now < _next:
                offsetx, offsety = 0, -offsety
            else:
                offsetx = 0
        else:
            if _now < _next:
                offsetx, offsety = -offsetx, 0
            else:
                offsety = 0

        pnext = self.widget(_next).pos()
        pnow = self.widget(_now).pos()
        self.m_pnow = pnow

        offset = QPoint(offsetx, offsety)
        self.widget(_next).move(pnext - offset)
        self.widget(_next).show()
        self.widget(_next).raise_()

        animgroup = QParallelAnimationGroup(self, finished=self.animationDoneSlot)

        for index, start, end in zip((_now, _next), (pnow, pnext - offset), (pnow + offset, pnext)):
            animation = QPropertyAnimation(
                self.widget(index),
                b"pos",
                duration=self.m_speed,
                easingCurve=self.m_animationtype,
                startValue=start,
                endValue=end,
            )
            animgroup.addAnimation(animation)

        self.m_next = _next
        self.m_now = _now
        self.m_active = True
        animgroup.start(QAbstractAnimation.DeleteWhenStopped)

    @pyqtSlot()
    def animationDoneSlot(self):
        self.setCurrentIndex(self.m_next)
        self.widget(self.m_now).hide()
        self.widget(self.m_now).move(self.m_pnow)
        self.m_active = False

    def event(self, e: QEvent):
        if e.type() == QEvent.Gesture:
            return self.gestureEvent(QGestureEvent(e))
        return super(SlidingStackedWidget, self).event(e)

    def gestureEvent(self, e: QGestureEvent):
        if swipe := e.gesture(Qt.SwipeGesture):
            self.swipeTriggered(swipe)
        return True

    def swipeTriggered(self, g: QSwipeGesture):
        if g.state() == Qt.GestureFinished:
            if g.horizontalDirection() == QSwipeGesture.Left:
                self.slideInPrev()
            else:
                self.slideInNext()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/settings/widgets/rpc.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_RPCSettings(object):
    def setupUi(self, RPCSettings):
        RPCSettings.setObjectName("RPCSettings")
        RPCSettings.resize(174, 146)
        RPCSettings.setWindowTitle("DiscordRPC")
        self.layout = QtWidgets.QGridLayout(RPCSettings)
        self.layout.setObjectName("layout")
        self.enable = QtWidgets.QComboBox(RPCSettings)
        self.enable.setObjectName("enable")
        self.enable.addItem("")
        self.enable.addItem("")
        self.enable.addItem("")
        self.layout.addWidget(self.enable, 0, 1, 1, 1)
        self.label = QtWidgets.QLabel(RPCSettings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.layout.addWidget(self.label, 0, 0, 1, 1)
        self.show_game = QtWidgets.QCheckBox(RPCSettings)
        self.show_game.setObjectName("show_game")
        self.layout.addWidget(self.show_game, 1, 0, 1, 2)
        self.show_os = QtWidgets.QCheckBox(RPCSettings)
        self.show_os.setObjectName("show_os")
        self.layout.addWidget(self.show_os, 2, 0, 1, 2)
        self.show_time = QtWidgets.QCheckBox(RPCSettings)
        self.show_time.setObjectName("show_time")
        self.layout.addWidget(self.show_time, 3, 0, 1, 2)

        self.retranslateUi(RPCSettings)

    def retranslateUi(self, RPCSettings):
        _translate = QtCore.QCoreApplication.translate
        RPCSettings.setTitle(_translate("RPCSettings", "Discord RPC"))
        self.enable.setItemText(0, _translate("RPCSettings", "When Playing"))
        self.enable.setItemText(1, _translate("RPCSettings", "Always"))
        self.enable.setItemText(2, _translate("RPCSettings", "Never"))
        self.label.setText(_translate("RPCSettings", "Show"))
        self.show_game.setText(_translate("RPCSettings", "Show Game"))
        self.show_os.setText(_translate("RPCSettings", "Show OS"))
        self.show_time.setText(_translate("RPCSettings", "Show Time playing"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    RPCSettings = QtWidgets.QGroupBox()
    ui = Ui_RPCSettings()
    ui.setupUi(RPCSettings)
    RPCSettings.show()
    sys.exit(app.exec_())



import functools
import logging
import os
import queue
import subprocess
import time
from typing import Optional, Union, Tuple

from legendary.cli import LegendaryCLI as LegendaryCLIReal
from legendary.lfs.wine_helpers import case_insensitive_file_search
from legendary.models.downloading import AnalysisResult, ConditionCheckResult
from legendary.models.game import Game, InstalledGame, VerifyResult
from legendary.lfs.utils import validate_files
from legendary.utils.selective_dl import get_sdl_appname

from rare.lgndr.core import LegendaryCore
from rare.lgndr.downloader.mp.manager import DLManager
from rare.lgndr.glue.arguments import (
    LgndrInstallGameArgs,
    LgndrImportGameArgs,
    LgndrVerifyGameArgs,
    LgndrUninstallGameArgs,
    LgndrInstallGameRealArgs,
    LgndrInstallGameRealRet,
)
from rare.lgndr.glue.monkeys import LgndrIndirectStatus, LgndrIndirectLogger


# fmt: off
class LegendaryCLI(LegendaryCLIReal):

    # noinspection PyMissingConstructor
    def __init__(self, core: LegendaryCore):
        self.core = core
        self.logger = logging.getLogger('cli')
        self.logging_queue = None
        self.ql = self.setup_threaded_logging()

    def __del__(self):
        self.ql.stop()

    @staticmethod
    def unlock_installed(func):
        @functools.wraps(func)
        def unlock(self, *args, **kwargs):
            self.logger.debug("%s: Using unlock decorator", func.__name__)
            if not self.core.lgd.lock_installed():
                self.logger.info("Data is locked, trying to forcefully release it")
                # self.core.lgd._installed_lock.release(force=True)
            try:
                ret = func(self, *args, **kwargs)
            except Exception as e:
                raise e
            finally:
                self.core.lgd.unlock_installed()
            return ret
        return unlock

    def resolve_aliases(self, name):
        return super(LegendaryCLI, self)._resolve_aliases(name)

    @unlock_installed.__func__
    def install_game(self, args: LgndrInstallGameArgs) -> Optional[Tuple[DLManager, AnalysisResult, InstalledGame, Game, bool, Optional[str], ConditionCheckResult]]:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger)
        get_boolean_choice = args.get_boolean_choice
        sdl_prompt = args.sdl_prompt
        if not self.core.lgd.lock_installed():
            logger.fatal('Failed to acquire installed data lock, only one instance of Legendary may '
                         'install/import/move applications at a time.')
            return

        args.app_name = self._resolve_aliases(args.app_name)
        if self.core.is_installed(args.app_name):
            igame = self.core.get_installed_game(args.app_name)
            args.platform = igame.platform
            if igame.needs_verification and not args.repair_mode:
                logger.info('Game needs to be verified before updating, switching to repair mode...')
                args.repair_mode = True

        repair_file = None
        # Rare: The 'args.no_install' flags is set externally from the InstallDialog
        if args.repair_mode:
            args.repair_mode = True
            args.no_install = args.repair_and_update is False
            repair_file = os.path.join(self.core.lgd.get_tmp_path(), f'{args.app_name}.repair')

        # Rare: Rare is already logged in

        if args.file_prefix or args.file_exclude_prefix:
            args.no_install = True

        # Rare: Rare runs updates on already installed games only

        game = self.core.get_game(args.app_name, update_meta=True, platform=args.platform)

        if not game:
            logger.error(f'Could not find "{args.app_name}" in list of available games, '
                         f'did you type the name correctly?')
            return

        # Rare: Rare checks this before calling 'install_game'

        if args.platform not in game.asset_infos:
            if not args.no_install:
                if self.core.lgd.config.getboolean('Legendary', 'install_platform_fallback', fallback=True):
                    logger.warning(f'App has no asset for platform "{args.platform}", falling back to "Windows".')
                    args.platform = 'Windows'
                else:
                    logger.error(f'No app asset found for platform "{args.platform}", run '
                                 f'"legendary info --platform {args.platform}" and make '
                                 f'sure the app is available for the specified platform.')
                    return
            else:
                logger.warning(f'No asset found for platform "{args.platform}", '
                               f'trying anyway since --no-install is set.')

        if game.is_dlc:
            logger.info('Install candidate is DLC')
            app_name = game.metadata['mainGameItem']['releaseInfo'][0]['appId']
            base_game = self.core.get_game(app_name)
            # check if base_game is actually installed
            if not self.core.is_installed(app_name):
                # download mode doesn't care about whether something's installed
                if not args.no_install:
                    logger.fatal(f'Base game "{app_name}" is not installed!')
                    return
        else:
            base_game = None

        if args.repair_mode:
            if not self.core.is_installed(game.app_name):
                logger.error(f'Game "{game.app_title}" ({game.app_name}) is not installed!')
                return

            if not os.path.exists(repair_file):
                logger.info('Game has not been verified yet.')
                # Rare: we do not want to verify while preparing the download in the InstallDialog
                # Rare: we handle it differently through the GameInfo tab
                logger.error('Game has not been verified yet.')
                return
            else:
                logger.info(f'Using existing repair file: {repair_file}')

        # check if SDL should be disabled
        sdl_enabled = not args.install_tag and not game.is_dlc
        config_tags = self.core.lgd.config.get(game.app_name, 'install_tags', fallback=None)
        config_disable_sdl = self.core.lgd.config.getboolean(game.app_name, 'disable_sdl', fallback=False)
        # remove config flag if SDL is reset
        if config_disable_sdl and args.reset_sdl and not args.disable_sdl:
            self.core.lgd.config.remove_option(game.app_name, 'disable_sdl')
        # if config flag is not yet set, set it and remove previous install tags
        elif not config_disable_sdl and args.disable_sdl:
            logger.info('Clearing install tags from config and disabling SDL for title.')
            if config_tags:
                self.core.lgd.config.remove_option(game.app_name, 'install_tags')
                config_tags = None
            self.core.lgd.config.set(game.app_name, 'disable_sdl', 'true')
            sdl_enabled = False
        # just disable SDL, but keep config tags that have been manually specified
        elif config_disable_sdl or args.disable_sdl:
            sdl_enabled = False

        if sdl_enabled and ((sdl_name := get_sdl_appname(game.app_name)) is not None):
            if not self.core.is_installed(game.app_name) or config_tags is None or args.reset_sdl:
                sdl_data = self.core.get_sdl_data(sdl_name, platform=args.platform)
                if sdl_data:
                    if args.skip_sdl:
                        args.install_tag = ['']
                        if '__required' in sdl_data:
                            args.install_tag.extend(sdl_data['__required']['tags'])
                    else:
                        args.install_tag = sdl_prompt(sdl_data, game.app_title)
                    self.core.lgd.config.set(game.app_name, 'install_tags', ','.join(args.install_tag))
                else:
                    logger.error(f'Unable to get SDL data for {sdl_name}')
            else:
                args.install_tag = config_tags.split(',')
        elif args.install_tag and not game.is_dlc and not args.no_install:
            config_tags = ','.join(args.install_tag)
            logger.info(f'Saving install tags for "{game.app_name}" to config: {config_tags}')
            self.core.lgd.config.set(game.app_name, 'install_tags', config_tags)
        elif not game.is_dlc:
            if config_tags and args.reset_sdl:
                logger.info('Clearing install tags from config.')
                self.core.lgd.config.remove_option(game.app_name, 'install_tags')
            elif config_tags:
                logger.info(f'Using install tags from config: {config_tags}')
                args.install_tag = config_tags.split(',')

        logger.info(f'Preparing download for "{game.app_title}" ({game.app_name})...')
        # todo use status queue to print progress from CLI
        # This has become a little ridiculous hasn't it?
        dlm, analysis, igame = self.core.prepare_download(game=game, base_game=base_game, base_path=args.base_path,
                                                          force=args.force, max_shm=args.shared_memory,
                                                          max_workers=args.max_workers, game_folder=args.game_folder,
                                                          disable_patching=args.disable_patching,
                                                          override_manifest=args.override_manifest,
                                                          override_old_manifest=args.override_old_manifest,
                                                          override_base_url=args.override_base_url,
                                                          platform=args.platform,
                                                          file_prefix_filter=args.file_prefix,
                                                          file_exclude_filter=args.file_exclude_prefix,
                                                          file_install_tag=args.install_tag,
                                                          dl_optimizations=args.order_opt,
                                                          dl_timeout=args.dl_timeout,
                                                          repair=args.repair_mode,
                                                          repair_use_latest=args.repair_and_update,
                                                          disable_delta=args.disable_delta,
                                                          override_delta_manifest=args.override_delta_manifest,
                                                          preferred_cdn=args.preferred_cdn,
                                                          disable_https=args.disable_https,
                                                          bind_ip=args.bind_ip)

        # game is either up-to-date or hasn't changed, so we have nothing to do
        if not analysis.dl_size and not game.is_dlc:
            logger.info('Download size is 0, the game is either already up to date or has not changed. Exiting...')
            self.install_game_cleanup(game, igame, args.repair_mode, repair_file)
            # return
            # Rare: Return what we know about the download to queue a 0 size DLC
            res = self.core.check_installation_conditions(analysis=analysis, install=igame, game=game,
                                                          updating=self.core.is_installed(args.app_name),
                                                          ignore_space_req=args.ignore_space)
            return dlm, analysis, igame, game, args.repair_mode, repair_file, res

        res = self.core.check_installation_conditions(analysis=analysis, install=igame, game=game,
                                                      updating=self.core.is_installed(args.app_name),
                                                      ignore_space_req=args.ignore_space)

        return dlm, analysis, igame, game, args.repair_mode, repair_file, res

    # Rare: This is currently handled in DownloadThread, this is a trial
    @unlock_installed.__func__
    def install_game_real(self, args: LgndrInstallGameRealArgs, dlm: DLManager, game: Game, igame: InstalledGame) -> LgndrInstallGameRealRet:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger)
        ret = LgndrInstallGameRealRet(game.app_name)
        if not self.core.lgd.lock_installed():
            logger.fatal('Failed to acquire installed data lock, only one instance of Legendary may '
                         'install/import/move applications at a time.')
            return ret

        start_t = time.time()

        try:
            # set up logging stuff (should be moved somewhere else later)
            dlm.logging_queue = self.logging_queue
            dlm.proc_debug = args.dlm_debug

            dlm.start()
            while dlm.is_alive():
                try:
                    args.ui_update(dlm.status_queue.get(timeout=1.0))
                except queue.Empty:
                    pass
                if args.dlm_signals.update:
                    try:
                        dlm.signals_queue.put(args.dlm_signals, block=False, timeout=1.0)
                    except queue.Full:
                        pass
                time.sleep(dlm.update_interval / 10)
            dlm.join()
        except Exception as e:
            end_t = time.time()
            logger.info(f'Installation failed after {end_t - start_t:.02f} seconds.')
            logger.warning(f'The following exception occurred while waiting for the downloader to finish: {e!r}. '
                           f'Try restarting the process, the resume file will be used to start where it failed. '
                           f'If it continues to fail please open an issue on GitHub.')
            ret.ret_code = ret.ReturnCode.ERROR
            ret.message = f"{e!r}"
            return ret
        else:
            end_t = time.time()
            if args.dlm_signals.kill is True:
                logger.info(f"Download stopped after {end_t - start_t:.02f} seconds.")
                ret.exit_code = ret.ReturnCode.STOPPED
                return ret
            logger.info(f"Download finished in {end_t - start_t:.02f} seconds.")
            if not args.no_install:
                # Allow setting savegame directory at install time so sync-saves will work immediately
                if (game.supports_cloud_saves or game.supports_mac_cloud_saves) and args.save_path:
                    igame.save_path = args.save_path

                postinstall = self.core.install_game(igame)
                if postinstall:
                    self._handle_postinstall(postinstall, igame, skip_prereqs=args.yes, choice=args.install_prereqs)

                dlcs = self.core.get_dlc_for_game(game.app_name)
                if dlcs and not args.skip_dlcs:
                    for dlc in dlcs:
                        ret.dlcs.append(
                            {
                                "app_name": dlc.app_name,
                                "app_title": dlc.app_title,
                                "app_version": dlc.app_version(args.platform)
                            }
                        )

                    # Rare: We do not install DLCs automatically, we offer to do so through our downloads tab

                if (game.supports_cloud_saves or game.supports_mac_cloud_saves) and not game.is_dlc:
                    # todo option to automatically download saves after the installation
                    #  args does not have the required attributes for sync_saves in here,
                    #  not sure how to solve that elegantly.
                    logger.info(f'This game supports cloud saves, syncing is handled by the "sync-saves" command. '
                                f'To download saves for this game run "legendary sync-saves {args.app_name}"')
                    ret.sync_saves = True

                # show tip again after installation finishes so users hopefully actually see it
                if tip_url := self.core.get_game_tip(igame.app_name):
                    ret.tip_url = tip_url

            self.install_game_cleanup(game, igame, args.repair_mode, args.repair_file)

            logger.info(f'Finished installation process in {end_t - start_t:.02f} seconds.')

            return ret

    @unlock_installed.__func__
    def install_game_cleanup(self, game: Game, igame: InstalledGame, repair_mode: bool = False, repair_file: str = '') -> None:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(LgndrIndirectStatus(), self.logger)
        if not self.core.lgd.lock_installed():
            logger.fatal('Failed to acquire installed data lock, only one instance of Legendary may '
                         'install/import/move applications at a time.')
            return

        old_igame = self.core.get_installed_game(game.app_name)
        if old_igame and repair_mode and os.path.exists(repair_file):
            if old_igame.needs_verification:
                old_igame.needs_verification = False
                self.core.install_game(old_igame)

            logger.debug('Removing repair file.')
            os.remove(repair_file)

        # check if install tags have changed, if they did; try deleting files that are no longer required.
        if old_igame and old_igame.install_tags != igame.install_tags:
            old_igame.install_tags = igame.install_tags
            logger.info('Deleting now untagged files.')
            self.core.uninstall_tag(old_igame)
            self.core.install_game(old_igame)

        if old_igame.install_tags:
            self.core.lgd.config.set(game.app_name, 'install_tags', ','.join(old_igame.install_tags))
            self.core.lgd.save_config()

        # check if the version changed, this can happen for DLC that gets a version bump with no actual file changes
        if old_igame and old_igame.version != igame.version:
            old_igame.version = igame.version
            self.core.install_game(old_igame)

    def _handle_postinstall(self, postinstall, igame, skip_prereqs=False, choice=True):
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(LgndrIndirectStatus(), self.logger)
        # noinspection PyShadowingBuiltins
        def print(x): self.logger.info(x) if x else None
        # noinspection PyShadowingBuiltins
        def input(x): return 'y' if choice else 'n'

        print('\nThis game lists the following prerequisites to be installed:')
        print(f'- {postinstall["name"]}: {" ".join((postinstall["path"], postinstall["args"]))}')
        print('')

        if os.name == 'nt':
            if skip_prereqs:
                c = 'n'  # we don't want to launch anything, just silent install.
            else:
                choice = input('Do you wish to install the prerequisites? ([y]es, [n]o, [i]gnore): ')
                c = choice.lower()[0]
                print('')

            if c == 'i':  # just set it to installed
                logger.info('Marking prerequisites as installed...')
                self.core.prereq_installed(igame.app_name)
            elif c == 'y':  # set to installed and launch installation
                logger.info('Launching prerequisite executable..')
                self.core.prereq_installed(igame.app_name)
                req_path, req_exec = os.path.split(postinstall['path'])
                work_dir = os.path.join(igame.install_path, req_path)
                fullpath = os.path.join(work_dir, req_exec)
                try:
                    p = subprocess.Popen([fullpath, postinstall['args']], cwd=work_dir, shell=True)
                    p.wait()
                except Exception as e:
                    logger.error(f'Failed to run prereq executable with: {e!r}')
        else:
            logger.info('Automatic installation not available on Linux.')

    @unlock_installed.__func__
    def uninstall_game(self, args: LgndrUninstallGameArgs) -> None:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger, logging.WARNING)
        get_boolean_choice = args.get_boolean_choice_main
        if not self.core.lgd.lock_installed():
            logger.fatal('Failed to acquire installed data lock, only one instance of Legendary may '
                         'install/import/move applications at a time.')
            return

        args.app_name = self._resolve_aliases(args.app_name)
        igame = self.core.get_installed_game(args.app_name)
        if not igame:
            logger.error(f'Game {args.app_name} not installed, cannot uninstall!')
            return

        if not args.yes:
            if not get_boolean_choice(f'Do you wish to uninstall "{igame.title}"?', default=False):
                return

        if os.name == 'nt' and igame.uninstaller and not args.skip_uninstaller:
            self._handle_uninstaller(igame, args.yes, args)

        try:
            if not igame.is_dlc:
                # Remove DLC first so directory is empty when game uninstall runs
                dlcs = self.core.get_dlc_for_game(igame.app_name)
                for dlc in dlcs:
                    if (idlc := self.core.get_installed_game(dlc.app_name)) is not None:
                        logger.info(f'Uninstalling DLC "{dlc.app_name}"...')
                        self.core.uninstall_game(idlc, delete_files=not args.keep_files)

            logger.info(f'Removing "{igame.title}" from "{igame.install_path}"...')
            self.core.uninstall_game(igame, delete_files=not args.keep_files,
                                     delete_root_directory=not igame.is_dlc)
            logger.info('Game has been uninstalled.')
            return
        except Exception as e:
            logger.warning(f'Removing game failed: {e!r}, please remove {igame.install_path} manually.')
            return

    def _handle_uninstaller(self, igame: InstalledGame, yes=False, args: LgndrUninstallGameArgs = None):
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger, logging.WARNING)
        get_boolean_choice = args.get_boolean_choice_handler
        # noinspection PyShadowingBuiltins
        def print(x): self.logger.info(x) if x else None

        uninstaller = igame.uninstaller

        print('\nThis game provides the following uninstaller:')
        print(f'- {uninstaller["path"]} {uninstaller["args"]}\n')
        
        if yes or get_boolean_choice('Do you wish to run the uninstaller?', default=True):
            logger.info('Running uninstaller...')
            req_path, req_exec = os.path.split(uninstaller['path'])
            work_dir = os.path.join(igame.install_path, req_path)
            fullpath = os.path.join(work_dir, req_exec)
            try:
                p = subprocess.Popen([fullpath, uninstaller['args']], cwd=work_dir, shell=True)
                p.wait()
            except Exception as e:
                logger.error(f'Failed to run uninstaller: {e!r}')

    def verify_game(self, args: Union[LgndrVerifyGameArgs, LgndrInstallGameArgs], print_command=True, repair_mode=False, repair_online=False) -> Optional[Tuple[int, int]]:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger)

        args.app_name = self._resolve_aliases(args.app_name)
        if not self.core.is_installed(args.app_name):
            logger.error(f'Game "{args.app_name}" is not installed')
            return

        logger.info(f'Loading installed manifest for "{args.app_name}"')
        igame = self.core.get_installed_game(args.app_name)
        if not os.path.exists(igame.install_path):
            logger.error(f'Install path "{igame.install_path}" does not exist, make sure all necessary mounts '
                         f'are available. If you previously deleted the game folder without uninstalling, run '
                         f'"legendary uninstall -y {igame.app_name}" and reinstall from scratch.')
            return

        manifest_data, _ = self.core.get_installed_manifest(args.app_name)
        if manifest_data is None:
            if repair_mode:
                if not repair_online:
                    logger.critical('No manifest could be loaded, the manifest file may be missing!')
                    raise ValueError('Local manifest is missing')

                logger.warning('No manifest could be loaded, the file may be missing. Downloading the latest manifest.')
                game = self.core.get_game(args.app_name, platform=igame.platform)
                manifest_data, _ = self.core.get_cdn_manifest(game, igame.platform)
                # Rare: Save the manifest if we downloaded it because it was missing
                self.core.lgd.save_manifest(game.app_name, manifest_data,
                                            version=self.core.load_manifest(manifest_data).meta.build_version,
                                            platform=igame.platform)
            else:
                logger.critical(f'Manifest appears to be missing! To repair, run "legendary repair '
                                f'{args.app_name} --repair-and-update", this will however redownload all files '
                                f'that do not match the latest manifest in their entirety.')
                return

        manifest = self.core.load_manifest(manifest_data)

        files = sorted(manifest.file_manifest_list.elements,
                       key=lambda a: a.filename.lower())

        # build list of hashes
        if (config_tags := self.core.lgd.config.get(args.app_name, 'install_tags', fallback=None)) is not None:
            install_tags = set(i.strip() for i in config_tags.split(','))
            file_list = [
                (f.filename, f.sha_hash.hex())
                for f in files
                if any(it in install_tags for it in f.install_tags) or not f.install_tags
            ]
        else:
            file_list = [(f.filename, f.sha_hash.hex()) for f in files]

        total = len(file_list)
        total_size = sum(manifest.file_manifest_list.get_file_by_path(fm[0]).file_size
                         for fm in file_list)
        num = processed = last_processed = 0
        speed = 0.0
        percentage = 0.0
        failed = []
        missing = []

        last_update = time.time()

        logger.info(f'Verifying "{igame.title}" version "{manifest.meta.build_version}"')
        repair_file = []
        for result, path, result_hash, bytes_read in validate_files(igame.install_path, file_list):
            processed += bytes_read
            percentage = (processed / total_size) * 100.0
            num += 1

            if (delta := ((current_time := time.time()) - last_update)) > 1:
                last_update = current_time
                speed = (processed - last_processed) / 1024 / 1024 / delta
                last_processed = processed

            if args.verify_stdout:
                args.verify_stdout(num, total, percentage, speed)

            if result == VerifyResult.HASH_MATCH:
                repair_file.append(f'{result_hash}:{path}')
                continue
            elif result == VerifyResult.HASH_MISMATCH:
                logger.error(f'File does not match hash: "{path}"')
                repair_file.append(f'{result_hash}:{path}')
                failed.append(path)
            elif result == VerifyResult.FILE_MISSING:
                logger.error(f'File is missing: "{path}"')
                missing.append(path)
            else:
                logger.error(f'Other failure (see log), treating file as missing: "{path}"')
                missing.append(path)

        if args.verify_stdout:
            args.verify_stdout(num, total, percentage, speed)

        # always write repair file, even if all match
        if repair_file:
            repair_filename = os.path.join(self.core.lgd.get_tmp_path(), f'{args.app_name}.repair')
            with open(repair_filename, 'w', encoding='utf-8') as f:
                f.write('\n'.join(repair_file))
            logger.debug(f'Written repair file to "{repair_filename}"')

        if not missing and not failed:
            logger.info('Verification finished successfully.')
            return 0, 0
        else:
            logger.error(f'Verification failed, {len(failed)} file(s) corrupted, {len(missing)} file(s) are missing.')
            if print_command:
                logger.info(f'Run "legendary repair {args.app_name}" to repair your game installation.')
            return len(failed), len(missing)

    @unlock_installed.__func__
    def import_game(self, args: LgndrImportGameArgs) -> None:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger)
        get_boolean_choice = args.get_boolean_choice
        if not self.core.lgd.lock_installed():
            logger.fatal('Failed to acquire installed data lock, only one instance of Legendary may '
                         'install/import/move applications at a time.')
            return

        # make sure path is absolute
        args.app_path = os.path.abspath(args.app_path)
        args.app_name = self._resolve_aliases(args.app_name)

        if not os.path.exists(args.app_path):
            logger.error(f'Specified path "{args.app_path}" does not exist!')
            return

        if self.core.is_installed(args.app_name):
            logger.error('Game is already installed!')
            return

        if not self.core.login():
            logger.error('Log in failed!')
            return

        # do some basic checks
        game = self.core.get_game(args.app_name, update_meta=True, platform=args.platform)
        if not game:
            logger.fatal(f'Did not find game "{args.app_name}" on account.')
            return

        if game.is_dlc:
            release_info = game.metadata.get('mainGameItem', {}).get('releaseInfo')
            if release_info:
                main_game_appname = release_info[0]['appId']
                main_game_title = game.metadata['mainGameItem']['title']
                if not self.core.is_installed(main_game_appname):
                    logger.error(f'Import candidate is DLC but base game "{main_game_title}" '
                                 f'(App name: "{main_game_appname}") is not installed!')
                    return
            else:
                logger.fatal(f'Unable to get base game information for DLC, cannot continue.')
                return

        # get everything needed for import from core, then run additional checks.
        manifest, igame = self.core.import_game(game, args.app_path, platform=args.platform)
        exe_path = os.path.join(args.app_path, manifest.meta.launch_exe.lstrip('/'))
        if os.name != 'nt':
            exe_path = case_insensitive_file_search(exe_path)
        # check if most files at least exist or if user might have specified the wrong directory
        total = len(manifest.file_manifest_list.elements)
        found = sum(os.path.exists(os.path.join(args.app_path, f.filename))
                    for f in manifest.file_manifest_list.elements)
        ratio = found / total

        if not found:
            logger.error(f'No files belonging to {"DLC" if game.is_dlc else "Game"} "{game.app_title}" '
                         f'({game.app_name}) found in the specified location, please verify that the path is correct.')
            if not game.is_dlc:
                # check if game folder is in path, suggest alternative
                folder = game.metadata.get('customAttributes', {}).get('FolderName', {}).get('value', game.app_name)
                if folder and folder not in args.app_path:
                    new_path = os.path.join(args.app_path, folder)
                    logger.info(f'Did you mean "{new_path}"?')
            return

        if not game.is_dlc and not os.path.exists(exe_path) and not args.disable_check:
            logger.error(f'Game executable could not be found at "{exe_path}", '
                         f'please verify that the specified path is correct.')
            return

        if ratio < 0.95:
            logger.warning('Some files are missing from the game installation, install may not '
                           'match latest Epic Games Store version or might be corrupted.')
        else:
            logger.info(f'{"DLC" if game.is_dlc else "Game"} install appears to be complete.')

        self.core.install_game(igame)
        if igame.needs_verification:
            logger.info(f'NOTE: The {"DLC" if game.is_dlc else "Game"} installation will have to be '
                        f'verified before it can be updated with legendary.')
            logger.info(f'Run "legendary repair {args.app_name}" to do so.')
        else:
            logger.info(f'Installation had Epic Games Launcher metadata for version "{igame.version}", '
                        f'verification will not be required.')

        # check for importable DLC
        if not args.skip_dlcs:
            dlcs = self.core.get_dlc_for_game(game.app_name)
            if dlcs:
                logger.info(f'Found {len(dlcs)} items of DLC that could be imported.')
                import_dlc = True
                if not args.yes and not args.with_dlcs:
                    if not get_boolean_choice(f'Do you wish to automatically attempt to import all DLCs?'):
                        import_dlc = False

                if import_dlc:
                    for dlc in dlcs:
                        args.app_name = dlc.app_name
                        self.import_game(args)

        logger.info(f'{"DLC" if game.is_dlc else "Game"} "{game.app_title}" has been imported.')
        return

    @unlock_installed.__func__
    def egs_sync(self, args):
        return super(LegendaryCLI, self).egs_sync(args)

    @unlock_installed.__func__
    def move(self, args):
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger)
        if not self.core.lgd.lock_installed():
            logger.fatal('Failed to acquire installed data lock, only one instance of Legendary may '
                         'install/import/move applications at a time.')
            return

        app_name = self._resolve_aliases(args.app_name)
        igame = self.core.get_installed_game(app_name, skip_sync=True)
        if not igame:
            logger.error(f'No installed game found for "{app_name}"')
            return

        old_base, game_folder = os.path.split(igame.install_path.replace('\\', '/'))
        new_path = os.path.join(args.new_path, game_folder)
        logger.info(f'Moving "{game_folder}" from "{old_base}" to "{args.new_path}"')

        if not args.skip_move:
            try:
                if not os.path.exists(args.new_path):
                    os.makedirs(args.new_path)

                os.rename(igame.install_path, new_path)
            except Exception as e:
                # pylint: disable=E1101
                if isinstance(e, OSError) and e.errno == 18:
                    logger.error(f'Moving to a different drive is not supported. Move the folder manually to '
                                 f'"{new_path}" and run "legendary move {app_name} "{args.new_path}" --skip-move"')
                elif isinstance(e, FileExistsError):
                    logger.error(f'The target path already contains a folder called "{game_folder}", '
                                 f'please remove or rename it first.')
                else:
                    logger.error(f'Moving failed with unknown error {e!r}.')
                    logger.info(f'Try moving the folder manually to "{new_path}" and running '
                                f'"legendary move {app_name} "{args.new_path}" --skip-move"')
                return
        else:
            logger.info(f'Not moving, just rewriting legendary metadata...')

        igame.install_path = new_path
        self.core.install_game(igame)
        logger.info('Finished.')

# fmt: on



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QStackedWidget, QMessageBox

from rare.components.tabs.store import ShopApiCore
from rare.components.tabs.store.game_widgets import WishlistWidget
from rare.ui.components.tabs.store.wishlist import Ui_Wishlist
from rare.utils.extra_widgets import WaitingSpinner
from rare.utils.misc import icon


class Wishlist(QStackedWidget, Ui_Wishlist):
    show_game_info = pyqtSignal(dict)
    update_wishlist_signal = pyqtSignal()

    def __init__(self, api_core: ShopApiCore):
        super(Wishlist, self).__init__()
        self.api_core = api_core
        self.setupUi(self)
        self.addWidget(WaitingSpinner())
        self.setCurrentIndex(1)
        self.wishlist = []
        self.widgets = []

        self.sort_cb.currentIndexChanged.connect(
            lambda i: self.set_wishlist(self.wishlist, i)
        )
        self.filter_cb.currentIndexChanged.connect(self.set_filter)
        self.reload_button.clicked.connect(self.update_wishlist)
        self.reload_button.setIcon(icon("fa.refresh", color="white"))

        self.reverse.stateChanged.connect(
            lambda: self.set_wishlist(sort=self.sort_cb.currentIndex())
        )

    def update_wishlist(self):
        self.setCurrentIndex(1)
        self.api_core.get_wishlist(self.set_wishlist)

    def delete_from_wishlist(self, game):
        self.api_core.remove_from_wishlist(
            game["namespace"],
            game["id"],
            lambda success: self.update_wishlist()
            if success
            else QMessageBox.warning(
                self, "Error", self.tr("Could not remove game from wishlist")
            ),
        )
        self.update_wishlist_signal.emit()

    def set_filter(self, i):
        count = 0
        for w in self.widgets:
            if i == 1 and not w.discount:
                w.setVisible(False)
            else:
                w.setVisible(True)
                count += 1

            if i == 0:
                w.setVisible(True)

        if count == 0:
            self.no_games_label.setVisible(True)
        else:
            self.no_games_label.setVisible(False)

    def set_wishlist(self, wishlist=None, sort=0):
        if wishlist and wishlist[0] == "error":
            return

        if wishlist is not None:
            self.wishlist = wishlist

        for i in self.widgets:
            i.deleteLater()

        if sort == 0:
            sorted_list = sorted(self.wishlist, key=lambda x: x["offer"]["title"])
        elif sort == 1:
            sorted_list = sorted(
                self.wishlist,
                key=lambda x: x["offer"]["price"]["totalPrice"]["fmtPrice"][
                    "discountPrice"
                ],
            )
        elif sort == 2:
            sorted_list = sorted(
                self.wishlist, key=lambda x: x["offer"]["seller"]["name"]
            )
        elif sort == 3:
            sorted_list = sorted(
                self.wishlist,
                reverse=True,
                key=lambda x: 1
                - (
                    x["offer"]["price"]["totalPrice"]["discountPrice"]
                    / x["offer"]["price"]["totalPrice"]["originalPrice"]
                ),
            )
        else:
            sorted_list = self.wishlist
        self.widgets.clear()

        if len(sorted_list) == 0:
            self.no_games_label.setVisible(True)
        else:
            self.no_games_label.setVisible(False)

        if self.reverse.isChecked():
            sorted_list.reverse()

        for game in sorted_list:
            w = WishlistWidget(game["offer"])
            self.widgets.append(w)
            self.list_layout.addWidget(w)
            w.open_game.connect(self.show_game_info.emit)
            w.delete_from_wishlist.connect(self.delete_from_wishlist)
        self.setCurrentIndex(0)



import platform
import time
from argparse import Namespace
from enum import IntEnum
from logging import getLogger

from PyQt5.QtCore import QObject, pyqtSignal, QSettings
from requests.exceptions import ConnectionError, HTTPError

from rare.lgndr.core import LegendaryCore
from .worker import Worker

logger = getLogger("FetchWorker")


class FetchWorker(Worker):
    class Result(IntEnum):
        GAMES = 1
        NON_ASSET = 2
        COMBINED = 3

    class Signals(QObject):
        progress = pyqtSignal(int, str)
        result = pyqtSignal(object, int)

    def __init__(self, core: LegendaryCore, args: Namespace):
        super(Worker, self).__init__()
        self.signals = FetchWorker.Signals()
        self.core = core
        self.args = args
        self.settings = QSettings()

    def run_real(self):
        # Fetch regular EGL games with assets
        start_time = time.time()

        want_unreal = self.settings.value("unreal_meta", False, bool) or self.args.debug
        want_win32 = self.settings.value("win32_meta", False, bool)
        want_macos = self.settings.value("macos_meta", False, bool)
        need_macos = platform.system() == "Darwin"
        need_windows = not any([want_win32, want_macos, need_macos, self.args.debug])

        if want_win32 or self.args.debug:
            logger.info(
                "Requesting Win32 metadata due to %s, %s Unreal engine",
                "settings" if want_win32 else "debug",
                "with" if want_unreal else "without"
            )
            self.signals.progress.emit(00, self.signals.tr("Updating game metadata for Windows"))
            self.core.get_game_and_dlc_list(
                update_assets=not self.args.offline, platform="Win32", skip_ue=not want_unreal
            )

        if need_macos or want_macos or self.args.debug:
            logger.info(
                "Requesting macOS metadata due to %s, %s Unreal engine",
                "platform" if need_macos else "settings" if want_macos else "debug",
                "with" if want_unreal else "without"
            )
            self.signals.progress.emit(15, self.signals.tr("Updating game metadata for macOS"))
            self.core.get_game_and_dlc_list(
                update_assets=not self.args.offline, platform="Mac", skip_ue=not want_unreal
            )

        if need_windows:
            self.signals.progress.emit(00, self.signals.tr("Updating game metadata for Windows"))
            logger.info(
                "Requesting Windows metadata, %s Unreal engine",
                "with" if want_unreal else "without"
            )
        games, dlc_dict = self.core.get_game_and_dlc_list(
            update_assets=need_windows, platform="Windows", skip_ue=not want_unreal
        )
        logger.debug(f"Games {len(games)}, games with DLCs {len(dlc_dict)}")
        logger.debug(f"Request games: {time.time() - start_time} seconds")

        # Fetch non-asset games
        self.signals.progress.emit(30, self.signals.tr("Updating non-asset game metadata"))
        start_time = time.time()
        try:
            na_games, na_dlc_dict = self.core.get_non_asset_library_items(force_refresh=False, skip_ue=False)
        except (HTTPError, ConnectionError) as e:
            logger.warning(f"Exception while fetching non asset games from EGS: {e}")
            na_games, na_dlc_dict = ([], {})
        # FIXME:
        #  This is here because of broken appIds from Epic:
        #  https://discord.com/channels/826881530310819914/884510635642216499/1111321692703305729
        #  There is a tab character in the appId of Fallout New Vegas: Honest Hearts DLC, this breaks metadata storage
        #  on Windows as they can't handle tabs at the end of the filename (?)
        #  Legendary and Heroic are also affected, but it completely breaks Rare, so dodge it for now pending a fix.
        except Exception as e:
            logger.error(f"Exception while fetching non asset games from EGS: {e}")
            na_games, na_dlc_dict = ([], {})
        logger.debug(f"Non-asset {len(na_games)}, games with non-asset DLCs {len(na_dlc_dict)}")
        logger.debug(f"Request non-asset: {time.time() - start_time} seconds")

        # Combine the two games lists and the two dlc dictionaries between regular and non-asset results
        games += na_games
        for catalog_id, dlcs in na_dlc_dict.items():
            if catalog_id in dlc_dict.keys():
                dlc_dict[catalog_id] += dlcs
            else:
                dlc_dict[catalog_id] = dlcs
        logger.debug(f"Games {len(games)}, games with DLCs {len(dlc_dict)}")

        self.signals.result.emit((games, dlc_dict), FetchWorker.Result.COMBINED)



from PyQt5.QtGui import QShowEvent, QHideEvent
from PyQt5.QtWidgets import QStackedWidget, QTabWidget
from legendary.core import LegendaryCore

from rare.shared.rare_core import RareCore
from rare.utils.paths import cache_dir
from .game_info import ShopGameInfo
from .search_results import SearchResults
from .shop_api_core import ShopApiCore
from .shop_widget import ShopWidget
from .wishlist import WishlistWidget, Wishlist


class Shop(QStackedWidget):
    init = False

    def __init__(self, core: LegendaryCore):
        super(Shop, self).__init__()
        self.core = core
        self.rcore = RareCore.instance()
        self.api_core = ShopApiCore(
            self.core.egs.session.headers["Authorization"],
            self.core.language_code,
            self.core.country_code,
        )

        self.shop = ShopWidget(cache_dir(), self.core, self.api_core)
        self.wishlist_widget = Wishlist(self.api_core)

        self.store_tabs = QTabWidget(parent=self)
        self.store_tabs.addTab(self.shop, self.tr("Games"))
        self.store_tabs.addTab(self.wishlist_widget, self.tr("Wishlist"))

        self.addWidget(self.store_tabs)

        self.search_results = SearchResults(self.api_core)
        self.addWidget(self.search_results)
        self.search_results.show_info.connect(self.show_game_info)
        self.info = ShopGameInfo(
            [i.asset_infos["Windows"].namespace for i in self.rcore.game_list if bool(i.asset_infos)],
            self.api_core,
        )
        self.addWidget(self.info)
        self.info.back_button.clicked.connect(lambda: self.setCurrentIndex(0))

        self.search_results.back_button.clicked.connect(lambda: self.setCurrentIndex(0))
        self.shop.show_info.connect(self.show_search_results)

        self.wishlist_widget.show_game_info.connect(self.show_game_info)
        self.shop.show_game.connect(self.show_game_info)
        self.api_core.update_wishlist.connect(self.update_wishlist)
        self.wishlist_widget.update_wishlist_signal.connect(self.update_wishlist)

    def showEvent(self, a0: QShowEvent) -> None:
        if a0.spontaneous() or self.init:
            return super().showEvent(a0)
        self.shop.load()
        self.wishlist_widget.update_wishlist()
        self.init = True
        return super().showEvent(a0)

    def hideEvent(self, a0: QHideEvent) -> None:
        if a0.spontaneous():
            return super().hideEvent(a0)
        # TODO: Implement store unloading
        return super().hideEvent(a0)

    def update_wishlist(self):
        self.shop.update_wishlist()

    def show_game_info(self, data):
        self.info.update_game(data)
        self.setCurrentIndex(2)

    def show_search_results(self, text: str):
        self.search_results.load_results(text)
        self.setCurrentIndex(1)
